import functools
import itertools
from typing import List

import casadi as ca

import numpy as np

from .goal_programming_mixin import GoalProgrammingMixin
from .goal_programming_mixin_base import Goal, StateGoal, _GoalProgrammingMixinBase
from .goal_programming_mixin_base import _EmptyEnsembleList, _EmptyEnsembleOrderedDict, _GoalConstraint
from .single_pass_goal_programming_mixin import SinglePassGoalProgrammingMixin
from .timeseries import Timeseries


class MinAbsGoal(Goal):
    """
    Absolute minimization goal class which can be used to minimize the
    absolute value of the goal's (linear) goal function. Contrary to its super
    class, the default order is 1 as absolute minimization is typically
    desired for fully linear problems.
    """
    order = 1


class MinAbsStateGoal(StateGoal, MinAbsGoal):
    pass


class _ConvertedMinAbsGoal(Goal):

    order = 1

    def __init__(self, abs_variable, is_path_goal, orig_goal):
        self.abs_variable = abs_variable
        self.is_path_goal = is_path_goal
        self.orig_goal = orig_goal

        # Copy relevant properties
        self.size = orig_goal.size
        self.weight = orig_goal.weight
        self.relaxation = orig_goal.relaxation / orig_goal.function_nominal
        self.priority = orig_goal.priority

    def function(self, optimization_problem, ensemble_member):
        if self.is_path_goal:
            return optimization_problem.variable(self.abs_variable.name())
        else:
            return optimization_problem.extra_variable(self.abs_variable.name(), ensemble_member)


class MinAbsGoalProgrammingMixin(_GoalProgrammingMixinBase):
    """
    Similar behavior to :py:class:`.GoalProgrammingMixin`, but any
    :py:class:`MinAbsGoal` passed to :py:meth:`.min_abs_goals` or
    :py:meth:`.min_abs_path_goals` will be automatically converted to:

      1. An auxiliary minimization variable
      2. Two additional linear constraints relating the auxiliary variable to the goal function
      3. A new goal (of a different type) minimizing the auxiliary variable
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # List for any absolute minimization goals
        self.__problem_constraints = _EmptyEnsembleList()
        self.__problem_vars = []
        self.__problem_path_constraints = _EmptyEnsembleList()
        self.__problem_path_vars = []
        self.__seeds = _EmptyEnsembleOrderedDict()
        self.__path_seeds = _EmptyEnsembleOrderedDict()

        self.__first_run = True

    @property
    def extra_variables(self):
        return super().extra_variables + self.__problem_vars

    @property
    def path_variables(self):
        return super().path_variables + self.__problem_path_vars

    def bounds(self):
        bounds = super().bounds()
        for abs_var in (self.__problem_vars + self.__problem_path_vars):
            bounds[abs_var.name()] = (0.0, np.inf)
        return bounds

    def seed(self, ensemble_member):
        seed = super().seed(ensemble_member)

        # Seed minimization variables of current priority (not those of
        # previous priorities, as those are handled by GoalProgrammingMixin).
        for abs_var, val in self.__seeds[ensemble_member].items():
            seed[abs_var] = val

        times = self.times()
        for abs_var, val in self.__path_seeds[ensemble_member].items():
            seed[abs_var] = Timeseries(times, val)

        return seed

    def constraints(self, ensemble_member):
        constraints = super().constraints(ensemble_member)

        for constraint in self.__problem_constraints[ensemble_member]:
            constraints.append((constraint.function(self), constraint.min, constraint.max))

        return constraints

    def path_constraints(self, ensemble_member):
        path_constraints = super().path_constraints(ensemble_member)

        for constraint in self.__problem_path_constraints[ensemble_member]:
            path_constraints.append((constraint.function(self), constraint.min, constraint.max))

        return path_constraints

    def __validate_goals(self, goals, is_path_goal):
        goals = sorted(goals, key=lambda x: x.priority)

        for goal in goals:
            if not isinstance(goal, MinAbsGoal):
                raise Exception("Absolute goal not an instance of MinAbsGoal for goal {}".format(goal))

            if goal.function_range != (np.nan, np.nan):
                raise Exception("Absolute goal function is only allowed for minimization for goal {}".format(goal))

            if goal.order != 1:
                raise Exception("Absolute goal function is only allowed for order = 1 for goal {}".format(goal))

            if goal.weight <= 0:
                raise Exception("Absolute goal function is only allowed for weight > 0 for goal {}".format(goal))

    @staticmethod
    def __convert_goals(goals, sym_index, ensemble_size, is_path_goal):

        # Replace absolute minimization goals with a new goal, and some
        # additional hard constraints.
        constraints = [[] for ensemble_member in range(ensemble_size)]
        variables = []

        # It is easier to modify goals in place, but we do not want to modify
        # the original input list of goals. Make a copy to work with and
        # return when we are done.
        goals = goals.copy()

        for j, goal in enumerate(goals):
            assert isinstance(goal, MinAbsGoal)

            abs_variable_name = "abs_{}_{}".format(sym_index, j)
            if is_path_goal:
                abs_variable_name = "path_" + abs_variable_name

            abs_variable = ca.MX.sym(abs_variable_name, goal.size)
            variables.append(abs_variable)

            # Set constraints on how the additional variable relates to the
            # original goal function, such that it corresponds to its absolute
            # value when minimizing.
            for ensemble_member in range(ensemble_size):
                def _constraint_func(problem, sign, abs_variable=abs_variable,
                                     ensemble_member=ensemble_member, goal=goal,
                                     is_path_goal=is_path_goal):
                    if is_path_goal:
                        abs_variable = problem.variable(abs_variable.name())
                    else:
                        abs_variable = problem.extra_variable(abs_variable.name(), ensemble_member)

                    return abs_variable + sign * goal.function(problem, ensemble_member) / goal.function_nominal

                _pos = functools.partial(_constraint_func, sign=1)
                _neg = functools.partial(_constraint_func, sign=-1)

                constraints[ensemble_member].append(_GoalConstraint(None, _pos, 0.0, np.inf, False))
                constraints[ensemble_member].append(_GoalConstraint(None, _neg, 0.0, np.inf, False))

            # Overwrite the original goal, such that it is just a minimization
            # of the additional variable.
            goals[j] = _ConvertedMinAbsGoal(abs_variable, is_path_goal, goal)

        return goals, constraints, variables

    def __calculate_seed(self, goals, is_path_goal):
        assert self.__first_run is False

        seed = [{} for ensemble_member in range(self.ensemble_size)]

        for goal in goals:

            assert isinstance(goal, _ConvertedMinAbsGoal)

            for ensemble_member in range(self.ensemble_size):
                if is_path_goal:
                    expr = self.map_path_expression(goal.orig_goal.function(self, ensemble_member), ensemble_member)
                else:
                    expr = goal.orig_goal.function(self, ensemble_member)

                function = ca.Function('f', [self.solver_input], [expr])
                value = np.array(function(self.solver_output))

                assert value.ndim == 2

                if goal.size == 1:
                    if is_path_goal:
                        value = value.ravel()
                    else:
                        value = value.item()

                seed[ensemble_member][goal.abs_variable.name()] = np.abs(value)

        return seed

    def optimize(self, preprocessing=True, **kwargs):
        # Do pre-processing
        if preprocessing:
            self.pre()

        goals = self.min_abs_goals()
        path_goals = self.min_abs_path_goals()

        # Validate goal definitions
        self.__validate_goals(goals, is_path_goal=False)
        self.__validate_goals(path_goals, is_path_goal=True)

        # List for absolute minimization goals. These will be incrementally
        # filled only just before we need them to.
        self.__problem_constraints = [[] for ensemble_member in range(self.ensemble_size)]
        self.__problem_vars = []
        self.__problem_path_constraints = [[] for ensemble_member in range(self.ensemble_size)]
        self.__problem_path_vars = []

        # Similar to the above, but these keep track of all auxiliary
        # variables and constraints of priorities we have had (and therefore
        # activated), and those yet to come (and have not yet activated).
        self.__subproblem_constraints = {}
        self.__subproblem_vars = {}
        self.__subproblem_abs_goals = {}
        self.__subproblem_path_constraints = {}
        self.__subproblem_path_vars = {}
        self.__subproblem_path_abs_goals = {}

        # We want to have consistent naming with GPMixin for our auxiliary
        # variables. We therefore need to loop over all priorities, regardless
        # of whether there are any MinAbsGoals in it or not.
        priorities = {int(goal.priority) for goal in itertools.chain(
            goals, path_goals, self.goals(), self.path_goals()) if not goal.is_empty}

        subproblems = []
        for priority in sorted(priorities):
            subproblems.append((
                priority,
                [goal for goal in goals if int(goal.priority) == priority and not goal.is_empty],
                [goal for goal in path_goals if int(goal.priority) == priority and not goal.is_empty]))

        # Rewrite absolute minimization goals.
        self.__converted_goals = []
        self.__converted_path_goals = []

        for i, (priority, goals, path_goals) in enumerate(subproblems):

            (goals,
                self.__subproblem_constraints[priority],
                self.__subproblem_vars[priority]) = self.__convert_goals(goals, i, self.ensemble_size, False)

            self.__converted_goals.extend(goals)
            self.__subproblem_abs_goals[priority] = goals

            (path_goals,
                self.__subproblem_path_constraints[priority],
                self.__subproblem_path_vars[priority]) = self.__convert_goals(path_goals, i, self.ensemble_size, True)

            self.__converted_path_goals.extend(path_goals)
            self.__subproblem_path_abs_goals[priority] = path_goals

        return super().optimize(**kwargs, preprocessing=False)

    def priority_started(self, priority):
        super().priority_started(priority)

        # Enable constraints and auxiliary variables that we need starting
        # from this priority when using GoalProgrammingMixin. When using
        # SinglePassGoalProgrammingMixin, we need to add all constraints from
        # the start.
        if isinstance(self, GoalProgrammingMixin):
            priorities = [priority]
        elif isinstance(self, SinglePassGoalProgrammingMixin):
            if self.__first_run:
                priorities = self.__subproblem_constraints.keys()
            else:
                priorities = []

        for p in priorities:
            for a, b in zip(self.__problem_constraints, self.__subproblem_constraints[p]):
                a.extend(b)

            self.__problem_vars.extend(self.__subproblem_vars[p])

            for a, b in zip(self.__problem_path_constraints, self.__subproblem_path_constraints[p]):
                a.extend(b)

            self.__problem_path_vars.extend(self.__subproblem_path_vars[p])

        # Calculate the seed needed for goals/variables introduced in this
        # priority. We can only calculate a seed if this is not the first
        # priority.
        if not self.__first_run and isinstance(self, GoalProgrammingMixin):
            self.__seeds = self.__calculate_seed(self.__subproblem_abs_goals[priority], False)
            self.__path_seeds = self.__calculate_seed(self.__subproblem_path_abs_goals[priority], True)

        self.__first_run = False

    def min_abs_goals(self) -> List[MinAbsGoal]:
        """
        User problem returns list of :py:class:`MinAbsGoal` objects.

        :returns: A list of goals.
        """
        return []

    def goals(self):
        goals = super().goals()
        try:
            return goals + self.__converted_goals
        except AttributeError:
            return goals

    def min_abs_path_goals(self) -> List[MinAbsGoal]:
        """
        User problem returns list of :py:class:`MinAbsGoal` objects.

        :returns: A list of goals.
        """
        return []

    def path_goals(self):
        goals = super().path_goals()
        try:
            return goals + self.__converted_path_goals
        except AttributeError:
            return goals



import logging
from datetime import timedelta

import numpy as np

import rtctools.data.pi as pi
import rtctools.data.rtc as rtc
from rtctools.simulation.io_mixin import IOMixin

logger = logging.getLogger("rtctools")


class PIMixin(IOMixin):
    """
    Adds `Delft-FEWS Published Interface
    <https://publicwiki.deltares.nl/display/FEWSDOC/The+Delft-Fews+Published+Interface>`_
    I/O to your simulation problem.

    During preprocessing, files named ``rtcDataConfig.xml``, ``timeseries_import.xml``,  and``rtcParameterConfig.xml``
    are read from the ``input`` subfolder.  ``rtcDataConfig.xml`` maps
    tuples of FEWS identifiers, including location and parameter ID, to RTC-Tools time series identifiers.

    During postprocessing, a file named ``timeseries_export.xml`` is written to the ``output`` subfolder.

    :cvar pi_binary_timeseries: Whether to use PI binary timeseries format.  Default is ``False``.
    :cvar pi_parameter_config_basenames:
        List of parameter config file basenames to read. Default is [``rtcParameterConfig``].
    :cvar pi_check_for_duplicate_parameters: Check if duplicate parameters are read. Default is ``True``.
    :cvar pi_validate_timeseries: Check consistency of timeseries.  Default is ``True``.
    """

    #: Whether to use PI binary timeseries format
    pi_binary_timeseries = False

    #: Location of rtcParameterConfig files
    pi_parameter_config_basenames = ['rtcParameterConfig']

    #: Check consistency of timeseries
    pi_validate_timeseries = True

    #: Check for duplicate parameters
    pi_check_for_duplicate_parameters = True

    #: Ensemble member to read from input
    pi_ensemble_member = 0

    def __init__(self, **kwargs):
        # Call parent class first for default behaviour.
        super().__init__(**kwargs)

        # Load rtcDataConfig.xml.  We assume this file does not change over the
        # life time of this object.
        self.__data_config = rtc.DataConfig(self._input_folder)

    def read(self):
        # Call parent class first for default behaviour.
        super().read()

        # rtcParameterConfig
        self.__parameter_config = []
        try:
            for pi_parameter_config_basename in self.pi_parameter_config_basenames:
                self.__parameter_config.append(pi.ParameterConfig(
                    self._input_folder, pi_parameter_config_basename))
        except FileNotFoundError:
            raise FileNotFoundError(
                "PIMixin: {}.xml not found in {}.".format(pi_parameter_config_basename, self._input_folder))

        # Make a parameters dict for later access
        for parameter_config in self.__parameter_config:
            for location_id, model_id, parameter_id, value in parameter_config:
                try:
                    parameter = self.__data_config.parameter(parameter_id, location_id, model_id)
                except KeyError:
                    parameter = parameter_id
                self.io.set_parameter(parameter, value)

        try:
            self.__timeseries_import = pi.Timeseries(
                self.__data_config, self._input_folder, self.timeseries_import_basename,
                binary=self.pi_binary_timeseries, pi_validate_times=self.pi_validate_timeseries)
        except FileNotFoundError:
            raise FileNotFoundError('PIMixin: {}.xml not found in {}'.format(
                self.timeseries_import_basename, self._input_folder))

        self.__timeseries_export = pi.Timeseries(
            self.__data_config, self._output_folder, self.timeseries_export_basename,
            binary=self.pi_binary_timeseries, pi_validate_times=False, make_new_file=True)

        # Convert timeseries timestamps to seconds since t0 for internal use
        timeseries_import_times = self.__timeseries_import.times

        # Timestamp check
        if self.pi_validate_timeseries:
            for i in range(len(timeseries_import_times) - 1):
                if timeseries_import_times[i] >= timeseries_import_times[i + 1]:
                    raise ValueError(
                        'PIMixin: Time stamps must be strictly increasing.')

        # Check if the timeseries are equidistant
        dt = timeseries_import_times[1] - timeseries_import_times[0]
        if self.pi_validate_timeseries:
            for i in range(len(timeseries_import_times) - 1):
                if timeseries_import_times[i + 1] - timeseries_import_times[i] != dt:
                    raise ValueError(
                        'PIMixin: Expecting equidistant timeseries, the time step '
                        'towards {} is not the same as the time step(s) before. Set '
                        'unit to nonequidistant if this is intended.'.format(
                            timeseries_import_times[i + 1]))

        # Stick timeseries into an AliasDict
        self.io.reference_datetime = self.__timeseries_import.forecast_datetime

        debug = logger.getEffectiveLevel() == logging.DEBUG
        for variable, values in self.__timeseries_import.items(self.pi_ensemble_member):
            self.io.set_timeseries(variable, timeseries_import_times, values)
            if debug and variable in self.get_variables():
                logger.debug('PIMixin: Timeseries {} replaced another aliased timeseries.'.format(variable))

    def write(self):
        # Call parent class first for default behaviour.
        super().write()

        times = self._simulation_times
        if len(set(np.diff(times))) == 1:
            dt = timedelta(seconds=times[1] - times[0])
        else:
            dt = None

        # Start of write output
        # Write the time range for the export file.
        self.__timeseries_export.times = [self.io.reference_datetime + timedelta(seconds=s) for s in times]

        # Write other time settings
        self.__timeseries_export.forecast_datetime = self.io.reference_datetime
        self.__timeseries_export.dt = dt
        self.__timeseries_export.timezone = self.__timeseries_import.timezone

        # Write the ensemble properties for the export file.
        self.__timeseries_export.ensemble_size = 1
        self.__timeseries_export.contains_ensemble = self.__timeseries_import.contains_ensemble

        # For all variables that are output variables the values are
        # extracted from the results.
        for variable in self._io_output_variables:
            values = np.array(self._io_output[variable])
            # Check if ID mapping is present
            try:
                self.__data_config.pi_variable_ids(variable)
            except KeyError:
                logger.debug(
                    'PIMixin: variable {} has no mapping defined in rtcDataConfig '
                    'so cannot be added to the output file.'.format(variable))
                continue

            # Add series to output file
            self.__timeseries_export.set(variable, values, unit=self.__timeseries_import.get_unit(variable))

        # Write output file to disk
        self.__timeseries_export.write()

    @property
    def timeseries_import(self):
        """
        :class:`pi.Timeseries` object containing the input data.
        """
        return self.__timeseries_import

    @property
    def timeseries_import_times(self):
        """
        List of time stamps for which input data is specified.

        The time stamps are in seconds since t0, and may be negative.
        """
        return self.io.times_sec

    @property
    def timeseries_export(self):
        """
        :class:`pi.Timeseries` object for holding the output data.
        """
        return self.__timeseries_export

    def set_timeseries(self, variable, values, output=True, check_consistency=True, unit=None):
        if check_consistency:
            if len(self.times()) != len(values):
                raise ValueError(
                    'PIMixin: Trying to set/append values {} with a different '
                    'length than the forecast length. Please make sure the '
                    'values cover forecastDate through endDate with timestep {}.'.format(
                        variable, self.__timeseries_import.dt))

        if unit is None:
            unit = self.__timeseries_import.get_unit(variable)

        if output:
            try:
                self.__data_config.pi_variable_ids(variable)
            except KeyError:
                logger.debug(
                    'PIMixin: variable {} has no mapping defined in rtcDataConfig '
                    'so cannot be added to the output file.'.format(variable))
            else:
                self.__timeseries_export.set(variable, values, unit=unit)

        self.__timeseries_import.set(variable, values, unit=unit)
        self.io.set_timeseries(variable, self.io.datetimes, values)

    def get_timeseries(self, variable):
        _, values = self.io.get_timeseries(variable)
        return values

    def extract_results(self):
        """
        Extracts the results of output

        :returns: An AliasDict of output variables and results array format.
        """
        return self._io_output



from typing import Union

import casadi as ca

import numpy as np


class Timeseries:
    """
    Time series object, bundling time stamps with values.
    """

    def __init__(self, times: np.ndarray, values: Union[float, np.ndarray, list, ca.DM]):
        """
        Create a new time series object.

        :param times:  Iterable of time stamps.
        :param values: Iterable of values.
        """
        self.__times = times

        if isinstance(values, ca.DM):
            # Note that a ca.DM object has no __iter__ attribute, which we
            # want it to have. We also want it to store it as a _flat_ array,
            # not a 2-D column vector.
            assert values.shape[0] == 1 or values.shape[1] == 1, "Only 1D ca.DM objects supported"
            values = values.toarray().ravel()
        elif isinstance(values, (np.ndarray, list)) and len(values) == 1:
            values = values[0]

        if hasattr(values, '__iter__'):
            self.__values = np.array(values, dtype=np.float64, copy=True)
        else:
            self.__values = np.full_like(times, values, dtype=np.float64)

    @property
    def times(self) -> np.ndarray:
        """
        Array of time stamps.
        """
        return self.__times

    @property
    def values(self) -> np.ndarray:
        """
        Array of values.
        """
        return self.__values

    def __neg__(self) -> 'Timeseries':
        return self.__class__(self.times, -self.values)

    def __repr__(self) -> str:
        return 'Timeseries({}, {})'.format(self.__times, self.__values)



"""Toolbox for control and optimization of water systems.

RTC-Tools is the Deltares toolbox for control and optimization of water systems.

"""
from setuptools import setup, find_packages
import versioneer
import sys

if sys.version_info < (3, 5):
    sys.exit("Sorry, Python 3.5 or newer is required.")

DOCLINES = __doc__.split("\n")

CLASSIFIERS = """\
Development Status :: 4 - Beta
Intended Audience :: Science/Research
Intended Audience :: Information Technology
License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)
Programming Language :: Python
Programming Language :: Python :: 3
Topic :: Scientific/Engineering :: GIS
Topic :: Scientific/Engineering :: Mathematics
Topic :: Scientific/Engineering :: Physics
Operating System :: Microsoft :: Windows
Operating System :: POSIX
Operating System :: Unix
Operating System :: MacOS
"""

setup(
    name = 'rtc-tools',
    version = versioneer.get_version(),
    maintainer = 'Jorn Baayen',
    author = 'Jorn Baayen',
    description = DOCLINES[0],
    long_description = '\n'.join(DOCLINES[2:]),
    url = 'http://www.deltares.nl/en/software/rtc-tools/',
    download_url='http://gitlab.com/deltares/rtc-tools/',
    classifiers = [_f for _f in CLASSIFIERS.split('\n') if _f],
    platforms = ['Windows', 'Linux', 'Mac OS-X', 'Unix'],
    packages = find_packages("src"),
    package_dir = {"": "src"},
    install_requires = ["casadi >= 3.5.0",
                        "numpy >= 1.16.0",
                        "scipy >= 1.0.0",
                        "pymoca >= 0.4.1",
                        "rtc-tools-channel-flow >= 1.1.0a1"],
    tests_require = ['pytest', 'pytest-runner'],
    extras_require = {
        'netcdf':  ["netCDF4"],
        'all': ["netCDF4"],
    },
    python_requires='>=3.5',
    cmdclass = versioneer.get_cmdclass(),
    entry_points={
        'console_scripts': [
            'rtc-tools-download-examples = rtctools.rtctoolsapp:download_examples',
            'rtc-tools-copy-libraries = rtctools.rtctoolsapp:copy_libraries',
        ]
    },
)



from typing import List, Tuple, Union

from .goal_programming_mixin import Goal, GoalProgrammingMixin


class _MeasurementGoal(Goal):
    def __init__(self, state, measurement_id, max_deviation=1.0):
        self.__state = state
        self.__measurement_id = measurement_id

        self.function_nominal = max_deviation

    def function(self, optimization_problem, ensemble_member):
        op = optimization_problem
        return (
            op.state_at(self.__state, op.initial_time, ensemble_member) -
            op.timeseries_at(self.__measurement_id, op.initial_time, ensemble_member))

    order = 2
    priority = -2


class _SmoothingGoal(Goal):
    def __init__(self, state1, state2, max_deviation=1.0):
        self.__state1 = state1
        self.__state2 = state2

        self.function_nominal = max_deviation

    def function(self, optimization_problem, ensemble_member):
        op = optimization_problem
        return (
            op.state_at(self.__state1, op.initial_time, ensemble_member) -
            op.state_at(self.__state2, op.initial_time, ensemble_member))

    order = 2
    priority = -1


class InitialStateEstimationMixin(GoalProgrammingMixin):
    """
    Adds initial state estimation to your optimization problem *using goal programming*.

    Before any other goals are evaluated, first, the deviation between initial
    state measurements and their respective model states is minimized in the
    least squares sense (1DVAR, priority -2). Secondly, the distance between
    pairs of states is minimized, again in the least squares sense, so that
    "smooth" initial guesses are provided for states without measurements
    (priority -1).

    .. note::

        There are types of problems where, in addition to minimizing
        differences between states and measurements, it is advisable to
        perform a steady-state initialization using additional initial-time
        model equations.  For hydraulic models, for instance, it is often
        helpful to require that the time-derivative of the flow variables
        vanishes at the initial time.

    """

    def initial_state_measurements(self) -> List[Union[Tuple[str, str], Tuple[str, str, float]]]:
        """
        List of pairs ``(state, measurement_id)`` or triples ``(state, measurement_id, max_deviation)``,
        relating states to measurement time series IDs.

        The default maximum deviation is ``1.0``.
        """
        return []

    def initial_state_smoothing_pairs(self) -> List[Union[Tuple[str, str], Tuple[str, str, float]]]:
        """
        List of pairs ``(state1, state2)`` or triples ``(state1, state2, max_deviation)``, relating
        states the distance of which is to be minimized.

        The default maximum deviation is ``1.0``.
        """
        return []

    def goals(self):
        g = super().goals()

        for measurement in self.initial_state_measurements():
            g.append(_MeasurementGoal(*measurement))

        for smoothing_pair in self.initial_state_smoothing_pairs():
            g.append(_SmoothingGoal(*smoothing_pair))

        return g



import functools
import logging
import sys
from abc import ABCMeta, abstractmethod
from collections import OrderedDict
from typing import Callable, Dict, List, Union

import casadi as ca

import numpy as np

from .optimization_problem import OptimizationProblem
from .timeseries import Timeseries

logger = logging.getLogger("rtctools")


class _EmptyEnsembleList(list):
    """
    An indexable object containing infinitely many empty lists.
    Only to be used as a placeholder.
    """

    def __getitem__(self, key):
        return []


class _EmptyEnsembleOrderedDict(OrderedDict):
    """
    An indexable object containing infinitely many empty OrderedDicts.
    Only to be used as a placeholder.
    """

    def __getitem__(self, key):
        return OrderedDict()


class Goal(metaclass=ABCMeta):
    r"""
    Base class for lexicographic goal programming goals.

    A goal is defined by overriding the :func:`function` method.

    :cvar function_range:   Range of goal function.  *Required if a target is set*.
    :cvar function_nominal: Nominal value of function. Used for scaling.  Default is ``1``.
    :cvar target_min:       Desired lower bound for goal function.  Default is ``numpy.nan``.
    :cvar target_max:       Desired upper bound for goal function.  Default is ``numpy.nan``.
    :cvar priority:         Integer priority of goal.  Default is ``1``.
    :cvar weight:           Optional weighting applied to the goal.  Default is ``1.0``.
    :cvar order:            Penalization order of goal violation.  Default is ``2``.
    :cvar critical:         If ``True``, the algorithm will abort if this goal cannot be fully met.
                            Default is ``False``.
    :cvar relaxation:       Amount of slack added to the hard constraints related to the goal.
                            Must be a nonnegative value. Default is ``0.0``.

    The target bounds indicate the range within the function should stay, *if possible*.  Goals
    are, in that sense, *soft*, as opposed to standard hard constraints.

    Four types of goals can be created:

    1. Minimization goal if no target bounds are set:

       .. math::

            \min f

    2. Lower bound goal if ``target_min`` is set:

        .. math::

            m \leq f

    3. Upper bound goal if ``target_max`` is set:

        .. math::

            f \leq M

    4. Combined lower and upper bound goal if ``target_min`` and ``target_max`` are both set:

        .. math::

            m \leq f \leq M

    Lower priority goals take precedence over higher priority goals.

    Goals with the same priority are weighted off against each other in a single objective function.

    In goals where a target is set:
        * The function range interval must be provided as this is used to introduce hard constrains on the value that
          the function can take. If one is unsure about which value the function can take, it is recommended to
          overestimate this interval. However, an overestimated interval will negatively influence how accurately the
          target bounds are met.
        * The target provided must be contained in the function range.
        * The function nominal is used to scale the constraints.
        * If both a target_min and a target_max are set, the target maximum must be at least equal to minimum one.
        * In a path goal, the target can be a Timeseries.

    In minimization goals:
        * The function range is not used and therefore cannot be set.
        * The function nominal is used to scale the function value in the objective function. To ensure that all goals
          are given a similar importance, it is crucial to provide an accurate estimate of this parameter.

    The goal violation value is taken to the order'th power in the objective function of the final
    optimization problem.

    Relaxation is used to loosen the constraints that are set after the
    optimization of the goal's priority. The unit of the relaxation is equal
    to that of the goal function.

    A goal can be written in vector form. In a vector goal:
        * The goal size determines how many goals there are.
        * The goal function has shape ``(goal size, 1)``.
        * The function is either minimized or has, possibly various, targets.
        * Function nominal can either be an array with as many entries as the goal size or have a single value.
        * Function ranges can either be an array with as many entries as the goal size or have a single value.
        * In a goal, the target can either be an array with as many entries as the goal size or have a single value.
        * In a path goal, the target can also be a Timeseries whose values are either a 1-dimensional vector or have
          as many columns as the goal size.

    Example definition of the point goal :math:`x(t) \geq 1.1` for :math:`t=1.0` at priority 1::

        class MyGoal(Goal):
            def function(self, optimization_problem, ensemble_member):
                # State 'x' at time t = 1.0
                t = 1.0
                return optimization_problem.state_at('x', t, ensemble_member)

            function_range = (1.0, 2.0)
            target_min = 1.1
            priority = 1

    Example definition of the path goal :math:`x(t) \geq 1.1` for all :math:`t` at priority 2::

        class MyPathGoal(Goal):
            def function(self, optimization_problem, ensemble_member):
                # State 'x' at any point in time
                return optimization_problem.state('x')

            function_range = (1.0, 2.0)
            target_min = 1.1
            priority = 2

    Note that for path goals, the ensemble member index is not passed to the call
    to :func:`OptimizationProblem.state`.  This call returns a time-independent symbol
    that is also independent of the active ensemble member.  Path goals are
    applied to all times and all ensemble members simultaneously.

    """

    @abstractmethod
    def function(self, optimization_problem: OptimizationProblem, ensemble_member: int) -> ca.MX:
        """
        This method returns a CasADi :class:`MX` object describing the goal function.

        :returns: A CasADi :class:`MX` object.
        """
        pass

    #: Range of goal function
    function_range = (np.nan, np.nan)

    #: Nominal value of function (used for scaling)
    function_nominal = 1.0

    #: Desired lower bound for goal function
    target_min = np.nan

    #: Desired upper bound for goal function
    target_max = np.nan

    #: Lower priority goals take precedence over higher priority goals.
    priority = 1

    #: Goals with the same priority are weighted off against each other in a
    #: single objective function.
    weight = 1.0

    #: The goal violation value is taken to the order'th power in the objective
    #: function.
    order = 2

    #: The size of the goal if it's a vector goal.
    size = 1

    #: Critical goals must always be fully satisfied.
    critical = False

    #: Absolute relaxation applied to the optimized values of this goal
    relaxation = 0.0

    #: Timeseries ID for function value data (optional)
    function_value_timeseries_id = None

    #: Timeseries ID for goal violation data (optional)
    violation_timeseries_id = None

    @property
    def has_target_min(self) -> bool:
        """
        ``True`` if the user goal has min bounds.
        """
        if isinstance(self.target_min, Timeseries):
            return True
        else:
            return np.any(np.isfinite(self.target_min))

    @property
    def has_target_max(self) -> bool:
        """
        ``True`` if the user goal has max bounds.
        """
        if isinstance(self.target_max, Timeseries):
            return True
        else:
            return np.any(np.isfinite(self.target_max))

    @property
    def has_target_bounds(self) -> bool:
        """
        ``True`` if the user goal has min/max bounds.
        """
        return (self.has_target_min or self.has_target_max)

    @property
    def is_empty(self) -> bool:
        target_min_set = isinstance(self.target_min, Timeseries) or np.any(np.isfinite(self.target_min))
        target_max_set = isinstance(self.target_max, Timeseries) or np.any(np.isfinite(self.target_max))

        if not target_min_set and not target_max_set:
            # A minimization goal
            return False

        target_min = self.target_min
        if isinstance(target_min, Timeseries):
            target_min = target_min.values

        target_max = self.target_max
        if isinstance(target_max, Timeseries):
            target_max = target_max.values

        min_empty = not np.any(np.isfinite(target_min))
        max_empty = not np.any(np.isfinite(target_max))

        return min_empty and max_empty

    def get_function_key(self, optimization_problem: OptimizationProblem, ensemble_member: int) -> str:
        """
        Returns a key string uniquely identifying the goal function.  This
        is used to eliminate linearly dependent constraints from the optimization problem.
        """
        if hasattr(self, 'function_key'):
            return self.function_key

        # This must be deterministic.  See RTCTOOLS-485.
        if not hasattr(Goal, '_function_key_counter'):
            Goal._function_key_counter = 0
        self.function_key = '{}_{}'.format(self.__class__.__name__, Goal._function_key_counter)
        Goal._function_key_counter += 1

        return self.function_key

    def __repr__(self) -> str:
        return '{}(priority={}, target_min={}, target_max={}, function_range={})'.format(
            self.__class__, self.priority, self.target_min, self.target_max, self.function_range)


class StateGoal(Goal, metaclass=ABCMeta):
    r"""
    Base class for lexicographic goal programming path goals that act on a single model state.

    A state goal is defined by setting at least the ``state`` class variable.

    :cvar state:            State on which the goal acts.  *Required*.
    :cvar target_min:       Desired lower bound for goal function.  Default is ``numpy.nan``.
    :cvar target_max:       Desired upper bound for goal function.  Default is ``numpy.nan``.
    :cvar priority:         Integer priority of goal.  Default is ``1``.
    :cvar weight:           Optional weighting applied to the goal.  Default is ``1.0``.
    :cvar order:            Penalization order of goal violation.  Default is ``2``.
    :cvar critical:         If ``True``, the algorithm will abort if this goal cannot be fully met.
                            Default is ``False``.

    Example definition of the goal :math:`x(t) \geq 1.1` for all :math:`t` at priority 2::

        class MyStateGoal(StateGoal):
            state = 'x'
            target_min = 1.1
            priority = 2

    Contrary to ordinary ``Goal`` objects, ``PathGoal`` objects need to be initialized with an
    ``OptimizationProblem`` instance to allow extraction of state metadata, such as bounds and
    nominal values.  Consequently, state goals must be instantiated as follows::

        my_state_goal = MyStateGoal(optimization_problem)

    Note that ``StateGoal`` is a helper class.  State goals can also be defined using ``Goal`` as direct base class,
    by implementing the ``function`` method and providing the ``function_range`` and ``function_nominal``
    class variables manually.

    """

    #: The state on which the goal acts.
    state = None

    def __init__(self, optimization_problem):
        """
        Initialize the state goal object.

        :param optimization_problem: ``OptimizationProblem`` instance.
        """

        # Check whether a state has been specified
        if self.state is None:
            raise Exception('Please specify a state.')

        # Extract state range from model
        if self.has_target_bounds:
            try:
                self.function_range = optimization_problem.bounds()[self.state]
            except KeyError:
                raise Exception('State {} has no bounds or does not exist in the model.'.format(self.state))

            if self.function_range[0] is None:
                raise Exception('Please provide a lower bound for state {}.'.format(self.state))
            if self.function_range[1] is None:
                raise Exception('Please provide an upper bound for state {}.'.format(self.state))

        # Extract state nominal from model
        self.function_nominal = optimization_problem.variable_nominal(self.state)

        # Set function key
        canonical, sign = optimization_problem.alias_relation.canonical_signed(self.state)
        self.function_key = canonical if sign > 0.0 else '-' + canonical

    def function(self, optimization_problem, ensemble_member):
        return optimization_problem.state(self.state)

    def __repr__(self):
        return '{}(priority={}, state={}, target_min={}, target_max={}, function_range={})'.format(
            self.__class__, self.priority, self.state, self.target_min, self.target_max, self.function_range)


class _GoalConstraint:

    def __init__(
            self,
            goal: Goal,
            function: Callable[[OptimizationProblem], ca.MX],
            m: Union[float, np.ndarray, Timeseries],
            M: Union[float, np.ndarray, Timeseries],
            optimized: bool):

        assert isinstance(m, (float, np.ndarray, Timeseries))
        assert isinstance(M, (float, np.ndarray, Timeseries))
        assert type(m) == type(M)

        # NumPy arrays only allowed for vector goals
        if isinstance(m, np.ndarray):
            assert len(m) == goal.size
            assert len(M) == goal.size

        self.goal = goal
        self.function = function
        self.min = m
        self.max = M
        self.optimized = optimized

    def update_bounds(self, other, enforce='self'):
        # NOTE: a.update_bounds(b) is _not_ the same as  b.update_bounds(a).
        # See how the 'enforce' parameter is used.

        min_, max_ = self.min, self.max
        other_min, other_max = other.min, other.max

        if isinstance(min_, Timeseries):
            assert isinstance(max_, Timeseries)
            assert isinstance(other_min, Timeseries)
            assert isinstance(other_max, Timeseries)

            min_ = min_.values
            max_ = max_.values
            other_min = other_min.values
            other_max = other_max.values

        min_ = np.maximum(min_, other_min)
        max_ = np.minimum(max_, other_max)

        # Ensure new constraint bounds do not loosen or shift
        # previous bounds due to numerical errors.
        if enforce == 'self':
            min_ = np.minimum(max_, other_min)
            max_ = np.maximum(min_, other_max)
        else:
            min_ = np.minimum(min_, other_max)
            max_ = np.maximum(max_, other_min)

        # Ensure consistency of bounds. Bounds may become inconsistent due to
        # small numerical computation errors.
        min_ = np.minimum(min_, max_)

        if isinstance(self.min, Timeseries):
            self.min = Timeseries(self.min.times, min_)
            self.max = Timeseries(self.max.times, max_)
        else:
            self.min = min_
            self.max = max_


class _GoalProgrammingMixinBase(OptimizationProblem, metaclass=ABCMeta):

    def _gp_n_objectives(self, subproblem_objectives, subproblem_path_objectives, ensemble_member):
        return ca.vertcat(*[o(self, ensemble_member) for o in subproblem_objectives]).size1() \
               + ca.vertcat(*[o(self, ensemble_member) for o in subproblem_path_objectives]).size1()

    def _gp_objective(self, subproblem_objectives, n_objectives, ensemble_member):
        if len(subproblem_objectives) > 0:
            acc_objective = ca.sum1(ca.vertcat(*[o(self, ensemble_member) for o in subproblem_objectives]))

            if self.goal_programming_options()['scale_by_problem_size']:
                acc_objective = acc_objective / n_objectives

            return acc_objective
        else:
            return ca.MX(0)

    def _gp_path_objective(self, subproblem_path_objectives, n_objectives, ensemble_member):
        if len(subproblem_path_objectives) > 0:
            acc_objective = ca.sum1(ca.vertcat(*[o(self, ensemble_member) for o in subproblem_path_objectives]))

            if self.goal_programming_options()['scale_by_problem_size']:
                # Objective is already divided by number of active time steps
                # at this point when `scale_by_problem_size` is set.
                acc_objective = acc_objective / n_objectives

            return acc_objective
        else:
            return ca.MX(0)

    @abstractmethod
    def goal_programming_options(self) -> Dict[str, Union[float, bool]]:
        raise NotImplementedError()

    def goals(self) -> List[Goal]:
        """
        User problem returns list of :class:`Goal` objects.

        :returns: A list of goals.
        """
        return []

    def path_goals(self) -> List[Goal]:
        """
        User problem returns list of path :class:`Goal` objects.

        :returns: A list of path goals.
        """
        return []

    def _gp_min_max_arrays(self, g, target_shape=None):
        """
        Broadcasts the goal target minimum and target maximum to arrays of a desired target shape.

        Depending on whether g is a vector goal or not, the output shape differs:

        - A 2-D array of size (goal.size, target_shape or 1) if the goal size
          is larger than one, i.e. a vector goal
        - A 1-D array of size (target_shape or 1, ) otherwise
        """

        times = self.times()

        m, M = None, None
        if isinstance(g.target_min, Timeseries):
            m = self.interpolate(
                times, g.target_min.times, g.target_min.values, -np.inf, -np.inf)
            if m.ndim > 1:
                m = m.transpose()
        elif isinstance(g.target_min, np.ndarray) and target_shape:
            m = np.broadcast_to(g.target_min, (target_shape, g.size)).transpose()
        elif target_shape:
            m = np.full(target_shape, g.target_min)
        else:
            m = np.array([g.target_min]).transpose()
        if isinstance(g.target_max, Timeseries):
            M = self.interpolate(
                times, g.target_max.times, g.target_max.values, np.inf, np.inf)
            if M.ndim > 1:
                M = M.transpose()
        elif isinstance(g.target_max, np.ndarray) and target_shape:
            M = np.broadcast_to(g.target_max, (target_shape, g.size)).transpose()
        elif target_shape:
            M = np.full(target_shape, g.target_max)
        else:
            M = np.array([g.target_max]).transpose()

        if g.size > 1 and m.ndim == 1:
            m = np.broadcast_to(m, (g.size, len(m)))
        if g.size > 1 and M.ndim == 1:
            M = np.broadcast_to(M, (g.size, len(M)))

        if g.size > 1:
            assert m.shape == (g.size, 1 if target_shape is None else target_shape)
        else:
            assert m.shape == (1 if target_shape is None else target_shape, )
        assert m.shape == M.shape

        return m, M

    def _gp_validate_goals(self, goals, is_path_goal):
        goals = sorted(goals, key=lambda x: x.priority)

        options = self.goal_programming_options()

        # Validate goal definitions
        for goal in goals:
            m, M = goal.function_range

            # The function range should not be a symbolic expression
            if isinstance(m, ca.MX):
                assert m.is_constant()
                if m.size1() == 1:
                    m = float(m)
                else:
                    m = np.array(m.to_DM())

            if isinstance(M, ca.MX):
                assert M.is_constant()
                if M.size1() == 1:
                    M = float(M)
                else:
                    M = np.array(M.to_DM())

            assert isinstance(m, (float, int, np.ndarray))
            assert isinstance(M, (float, int, np.ndarray))

            if np.any(goal.function_nominal <= 0):
                raise Exception("Nonpositive nominal value specified for goal {}".format(goal))

            if goal.critical and not goal.has_target_bounds:
                raise Exception("Minimization goals cannot be critical")

            if goal.critical:
                # Allow a function range for backwards compatibility reasons.
                # Maybe raise a warning that its not actually used?
                pass
            elif goal.has_target_bounds:
                if not np.all(np.isfinite(m)) or not np.all(np.isfinite(M)):
                    raise Exception("No function range specified for goal {}".format(goal))

                if np.any(m >= M):
                    raise Exception("Invalid function range for goal {}".format(goal))

                if goal.weight <= 0:
                    raise Exception("Goal weight should be positive for goal {}".format(goal))
            else:
                if goal.function_range != (np.nan, np.nan):
                    raise Exception("Specifying function range not allowed for goal {}".format(goal))

            if not is_path_goal:
                if isinstance(goal.target_min, Timeseries):
                    raise Exception("Target min cannot be a Timeseries for goal {}".format(goal))
                if isinstance(goal.target_max, Timeseries):
                    raise Exception("Target max cannot be a Timeseries for goal {}".format(goal))

            try:
                int(goal.priority)
            except ValueError:
                raise Exception("Priority of not int or castable to int for goal {}".format(goal))

            if options['keep_soft_constraints']:
                if goal.relaxation != 0.0:
                    raise Exception("Relaxation not allowed with `keep_soft_constraints` for goal {}".format(goal))
                if goal.violation_timeseries_id is not None:
                    raise Exception(
                        "Violation timeseries id not allowed with `keep_soft_constraints` for goal {}".format(goal))
            else:
                if goal.size > 1:
                    raise Exception("Option `keep_soft_constraints` needs to be set for vector goal {}".format(goal))

            if goal.critical and goal.size > 1:
                raise Exception("Vector goal cannot be critical for goal {}".format(goal))

        if is_path_goal:
            target_shape = len(self.times())
        else:
            target_shape = None

        # Check consistency and monotonicity of goals. Scalar target min/max
        # of normal goals are also converted to arrays to unify checks with
        # path goals.
        if options['check_monotonicity']:
            for e in range(self.ensemble_size):
                # Store the previous goal of a certain function key we
                # encountered, such that we can compare to it.
                fk_goal_map = {}

                for goal in goals:
                    fk = goal.get_function_key(self, e)
                    prev = fk_goal_map.get(fk)
                    fk_goal_map[fk] = goal

                    if prev is not None:
                        goal_m, goal_M = self._gp_min_max_arrays(goal, target_shape)
                        other_m, other_M = self._gp_min_max_arrays(prev, target_shape)

                        indices = np.where(np.logical_not(np.logical_or(
                            np.isnan(goal_m), np.isnan(other_m))))
                        if goal.has_target_min:
                            if np.any(goal_m[indices] < other_m[indices]):
                                raise Exception(
                                    'Target minimum of goal {} must be greater or equal than '
                                    'target minimum of goal {}.'.format(goal, prev))

                        indices = np.where(np.logical_not(np.logical_or(
                            np.isnan(goal_M), np.isnan(other_M))))
                        if goal.has_target_max:
                            if np.any(goal_M[indices] > other_M[indices]):
                                raise Exception(
                                    'Target maximum of goal {} must be less or equal than '
                                    'target maximum of goal {}'.format(goal, prev))

        for goal in goals:
            goal_m, goal_M = self._gp_min_max_arrays(goal, target_shape)
            goal_lb = np.broadcast_to(goal.function_range[0], goal_m.shape[::-1]).transpose()
            goal_ub = np.broadcast_to(goal.function_range[1], goal_M.shape[::-1]).transpose()

            if goal.has_target_min and goal.has_target_max:
                indices = np.where(np.logical_not(np.logical_or(
                    np.isnan(goal_m), np.isnan(goal_M))))

                if np.any(goal_m[indices] > goal_M[indices]):
                    raise Exception("Target minimum exceeds target maximum for goal {}".format(goal))

            if goal.has_target_min and not goal.critical:
                indices = np.where(np.isfinite(goal_m))
                if np.any(goal_m[indices] <= goal_lb[indices]):
                    raise Exception(
                        'Target minimum should be greater than the lower bound of the function range for goal {}'
                        .format(goal))
                if np.any(goal_m[indices] > goal_ub[indices]):
                    raise Exception(
                        'Target minimum should not be greater than the upper bound of the function range for goal {}'
                        .format(goal))
            if goal.has_target_max and not goal.critical:
                indices = np.where(np.isfinite(goal_M))
                if np.any(goal_M[indices] >= goal_ub[indices]):
                    raise Exception(
                        'Target maximum should be smaller than the upper bound of the function range for goal {}'
                        .format(goal))
                if np.any(goal_M[indices] < goal_lb[indices]):
                    raise Exception(
                        'Target maximum should not be smaller than the lower bound of the function range for goal {}'
                        .format(goal))

            if goal.relaxation < 0.0:
                raise Exception('Relaxation of goal {} should be a nonnegative value'.format(goal))

    def _gp_goal_constraints(self, goals, sym_index, options, is_path_goal):
        """
        There are three ways in which a goal turns into objectives/constraints:

        1. A goal with target bounds results in a part for the objective (the
           violation variable), and 1 or 2 constraints (target min, max, or both).
        2. A goal without target bounds (i.e. minimization goal) results in just a
           part for the objective.
        3. A critical goal results in just a (pair of) constraint(s). These are hard
           constraints, which need to be put in the constraint store to guarantee
           linear independence.
        """

        epsilons = []
        objectives = []
        soft_constraints = [[] for ensemble_member in range(self.ensemble_size)]
        hard_constraints = [[] for ensemble_member in range(self.ensemble_size)]
        extra_constants = []

        eps_format = "eps_{}_{}"
        min_format = "min_{}_{}"
        max_format = "max_{}_{}"

        if is_path_goal:
            eps_format = "path_" + eps_format
            min_format = "path_" + min_format
            max_format = "path_" + max_format

        for j, goal in enumerate(goals):
            if goal.critical:
                assert goal.size == 1, "Critical goals cannot be vector goals"
                epsilon = np.zeros(len(self.times()) if is_path_goal else 1)
            elif goal.has_target_bounds:
                epsilon = ca.MX.sym(eps_format.format(sym_index, j), goal.size)
                epsilons.append(epsilon)

            # Make symbols for the target bounds (if set)
            if goal.has_target_min:
                min_variable = min_format.format(sym_index, j)

                # NOTE: When using a vector goal, we want to be sure that its constraints
                # and objective end up _exactly_ equal to its non-vector equivalent. We
                # therefore have to get rid of any superfluous/trivial constraints that
                # would otherwise be generated by the vector goal.
                target_min_slice_inds = np.full(goal.size, True)

                if isinstance(goal.target_min, Timeseries):
                    target_min = Timeseries(goal.target_min.times, goal.target_min.values)
                    inds = np.logical_or(np.isnan(target_min.values), np.isneginf(target_min.values))
                    target_min.values[inds] = -sys.float_info.max
                    n_times = len(goal.target_min.times)
                    target_min_slice_inds = ~np.all(np.broadcast_to(inds.transpose(), (goal.size, n_times)), axis=1)
                elif isinstance(goal.target_min, np.ndarray):
                    target_min = goal.target_min.copy()
                    inds = np.logical_or(np.isnan(target_min), np.isneginf(target_min))
                    target_min[inds] = -sys.float_info.max
                    target_min_slice_inds = ~inds
                else:
                    target_min = goal.target_min

                extra_constants.append((min_variable, target_min))
            else:
                min_variable = None

            if goal.has_target_max:
                max_variable = max_format.format(sym_index, j)

                target_max_slice_inds = np.full(goal.size, True)

                if isinstance(goal.target_max, Timeseries):
                    target_max = Timeseries(goal.target_max.times, goal.target_max.values)
                    inds = np.logical_or(np.isnan(target_max.values), np.isposinf(target_max.values))
                    target_max.values[inds] = sys.float_info.max
                    n_times = len(goal.target_max.times)
                    target_max_slice_inds = ~np.all(np.broadcast_to(inds.transpose(), (goal.size, n_times)), axis=1)
                elif isinstance(goal.target_max, np.ndarray):
                    target_max = goal.target_max.copy()
                    inds = np.logical_or(np.isnan(target_max), np.isposinf(target_max))
                    target_max[inds] = sys.float_info.max
                    target_max_slice_inds = ~inds
                else:
                    target_max = goal.target_max

                extra_constants.append((max_variable, target_max))
            else:
                max_variable = None

            # Make objective for soft constraints and minimization goals
            if not goal.critical:
                if hasattr(goal, '_objective_func'):
                    _objective_func = goal._objective_func
                elif goal.has_target_bounds:
                    if is_path_goal and options['scale_by_problem_size']:
                        goal_m, goal_M = self._gp_min_max_arrays(goal, target_shape=len(self.times()))
                        goal_active = np.isfinite(goal_m) | np.isfinite(goal_M)
                        n_active = np.sum(goal_active.astype(int), axis=-1)
                        # Avoid possible division by zero if goal is inactive
                        n_active = np.maximum(n_active, 1)
                    else:
                        n_active = 1

                    def _objective_func(problem, ensemble_member,
                                        goal=goal, epsilon=epsilon, is_path_goal=is_path_goal,
                                        n_active=n_active):
                        if is_path_goal:
                            epsilon = problem.variable(epsilon.name())
                        else:
                            epsilon = problem.extra_variable(epsilon.name(), ensemble_member)

                        return goal.weight * ca.constpow(epsilon, goal.order) / n_active
                else:
                    if is_path_goal and options['scale_by_problem_size']:
                        n_active = len(self.times())
                    else:
                        n_active = 1

                    def _objective_func(problem, ensemble_member, goal=goal, is_path_goal=is_path_goal,
                                        n_active=n_active):
                        f = goal.function(problem, ensemble_member) / goal.function_nominal
                        return goal.weight * ca.constpow(f, goal.order) / n_active

                objectives.append(_objective_func)

            # Make constraints for goals with target bounds
            if goal.has_target_bounds:
                if goal.critical:
                    for ensemble_member in range(self.ensemble_size):
                        constraint = self._gp_goal_hard_constraint(
                            goal, epsilon, None, ensemble_member, options, is_path_goal)
                        hard_constraints[ensemble_member].append(constraint)
                else:
                    for ensemble_member in range(self.ensemble_size):
                        # We use a violation variable formulation, with the violation
                        # variables epsilon bounded between 0 and 1.
                        def _soft_constraint_func(problem, target, bound, inds,
                                                  goal=goal, epsilon=epsilon, ensemble_member=ensemble_member,
                                                  is_path_constraint=is_path_goal):
                            if is_path_constraint:
                                target = problem.variable(target)
                                eps = problem.variable(epsilon.name())
                            else:
                                target = problem.parameters(ensemble_member)[target]
                                eps = problem.extra_variable(epsilon.name(), ensemble_member)

                            inds = inds.nonzero()[0].astype(int).tolist()

                            f = goal.function(problem, ensemble_member)
                            nominal = goal.function_nominal

                            return ca.if_else(ca.fabs(target) < sys.float_info.max,
                                              (f - eps * (bound - target) - target) / nominal,
                                              0.0)[inds]

                        if goal.has_target_min and np.any(target_min_slice_inds):
                            _f = functools.partial(
                                _soft_constraint_func,
                                target=min_variable,
                                bound=goal.function_range[0],
                                inds=target_min_slice_inds)
                            constraint = _GoalConstraint(goal, _f, 0.0, np.inf, False)
                            soft_constraints[ensemble_member].append(constraint)
                        if goal.has_target_max and np.any(target_max_slice_inds):
                            _f = functools.partial(
                                _soft_constraint_func,
                                target=max_variable,
                                bound=goal.function_range[1],
                                inds=target_max_slice_inds)
                            constraint = _GoalConstraint(goal, _f, -np.inf, 0.0, False)
                            soft_constraints[ensemble_member].append(constraint)

        return epsilons, objectives, soft_constraints, hard_constraints, extra_constants

    def _gp_goal_hard_constraint(self, goal, epsilon, existing_constraint, ensemble_member, options, is_path_goal):
        if not is_path_goal:
            epsilon = epsilon[:1]

        goal_m, goal_M = self._gp_min_max_arrays(goal, target_shape=epsilon.shape[0])

        if goal.has_target_bounds:
            # We use a violation variable formulation, with the violation
            # variables epsilon bounded between 0 and 1.
            m, M = np.full_like(epsilon, -np.inf, dtype=np.float64), np.full_like(epsilon, np.inf, dtype=np.float64)

            # A function range does not have to be specified for critical
            # goals. Avoid multiplying with NaN in that case.
            if goal.has_target_min:
                m = (epsilon * ((goal.function_range[0] - goal_m) if not goal.critical else 0.0)
                     + goal_m - goal.relaxation) / goal.function_nominal
            if goal.has_target_max:
                M = (epsilon * ((goal.function_range[1] - goal_M) if not goal.critical else 0.0)
                     + goal_M + goal.relaxation) / goal.function_nominal

            if goal.has_target_min and goal.has_target_max:
                # Avoid comparing with NaN
                inds = ~(np.isnan(m) | np.isnan(M))
                inds[inds] &= np.abs(m[inds] - M[inds]) < options['equality_threshold']
                if np.any(inds):
                    avg = 0.5 * (m + M)
                    m[inds] = M[inds] = avg[inds]

            m[~np.isfinite(goal_m)] = -np.inf
            M[~np.isfinite(goal_M)] = np.inf

            inds = epsilon > options['violation_tolerance']
            if np.any(inds):
                if is_path_goal:
                    expr = self.map_path_expression(goal.function(self, ensemble_member), ensemble_member)
                else:
                    expr = goal.function(self, ensemble_member)

                function = ca.Function('f', [self.solver_input], [expr])
                value = np.array(function(self.solver_output))

                m[inds] = (value - goal.relaxation) / goal.function_nominal
                M[inds] = (value + goal.relaxation) / goal.function_nominal

            m -= options['constraint_relaxation']
            M += options['constraint_relaxation']
        else:
            # Epsilon encodes the position within the function range.
            if options['fix_minimized_values'] and goal.relaxation == 0.0:
                m = epsilon / goal.function_nominal
                M = epsilon / goal.function_nominal
                self.check_collocation_linearity = False
                self.linear_collocation = False
            else:
                m = -np.inf * np.ones(epsilon.shape)
                M = (epsilon + goal.relaxation) / goal.function_nominal + options['constraint_relaxation']

        if is_path_goal:
            m = Timeseries(self.times(), m)
            M = Timeseries(self.times(), M)
        else:
            m = m[0]
            M = M[0]

        constraint = _GoalConstraint(
            goal,
            lambda problem, ensemble_member=ensemble_member, goal=goal: (
                goal.function(problem, ensemble_member) / goal.function_nominal),
            m, M, True)

        # Epsilon is fixed. Override previous {min,max} constraints for this
        # state.
        if existing_constraint:
            constraint.update_bounds(existing_constraint, enforce='other')

        return constraint

    def _gp_update_constraint_store(self, constraint_store, constraints):
        for ensemble_member in range(self.ensemble_size):
            for other in constraints[ensemble_member]:
                fk = other.goal.get_function_key(self, ensemble_member)
                try:
                    constraint_store[ensemble_member][fk].update_bounds(other)
                except KeyError:
                    constraint_store[ensemble_member][fk] = other

    def priority_started(self, priority: int) -> None:
        """
        Called when optimization for goals of certain priority is started.

        :param priority: The priority level that was started.
        """
        pass

    def priority_completed(self, priority: int) -> None:
        """
        Called after optimization for goals of certain priority is completed.

        :param priority: The priority level that was completed.
        """
        pass



import copy
import itertools
import logging
from collections import OrderedDict
from typing import Union

import casadi as ca

import numpy as np

import pkg_resources

import pymoca
import pymoca.backends.casadi.api


from rtctools._internal.alias_tools import AliasDict
from rtctools._internal.caching import cached
from rtctools._internal.debug_check_helpers import DebugLevel
from rtctools.data.storage import DataStoreAccessor

logger = logging.getLogger("rtctools")


class SimulationProblem(DataStoreAccessor):
    """
    Implements the `BMI <http://csdms.colorado.edu/wiki/BMI_Description>`_ Interface.

    Base class for all Simulation problems. Loads the Modelica Model.

    :cvar modelica_library_folders: Folders containing any referenced Modelica libraries. Default is an empty list.

    """

    _debug_check_level = DebugLevel.MEDIUM
    _debug_check_options = {}

    # Folders in which the referenced Modelica libraries are found
    modelica_library_folders = []

    # Force workaround for delay support by assuming zero delay. This flag
    # will be removed when proper delay support is added.
    _force_zero_delay = False

    def __init__(self, **kwargs):
        # Check arguments
        assert('model_folder' in kwargs)

        # Log pymoca version
        logger.debug("Using pymoca {}.".format(pymoca.__version__))

        # Transfer model from the Modelica .mo file to CasADi using pymoca
        if 'model_name' in kwargs:
            model_name = kwargs['model_name']
        else:
            if hasattr(self, 'model_name'):
                model_name = self.model_name
            else:
                model_name = self.__class__.__name__

        # Load model from pymoca backend
        self.__pymoca_model = pymoca.backends.casadi.api.transfer_model(
            kwargs['model_folder'], model_name, self.compiler_options())

        # Extract the CasADi MX variables used in the model
        self.__mx = {}
        self.__mx['time'] = [self.__pymoca_model.time]
        self.__mx['states'] = [v.symbol for v in self.__pymoca_model.states]
        self.__mx['derivatives'] = [v.symbol for v in self.__pymoca_model.der_states]
        self.__mx['algebraics'] = [v.symbol for v in self.__pymoca_model.alg_states]
        self.__mx['parameters'] = [v.symbol for v in self.__pymoca_model.parameters]
        self.__mx['constant_inputs'] = []
        self.__mx['lookup_tables'] = []

        for v in self.__pymoca_model.inputs:
            if v.symbol.name() in self.__pymoca_model.delay_states:
                # Delayed feedback variables are local to each ensemble, and
                # therefore belong to the collection of algebraic variables,
                # rather than to the control inputs.
                self.__mx['algebraics'].append(v.symbol)
            else:
                if v.symbol.name() in kwargs.get('lookup_tables', []):
                    self.__mx['lookup_tables'].append(v.symbol)
                else:
                    # All inputs are constant inputs
                    self.__mx['constant_inputs'].append(v.symbol)

        # Log variables in debug mode
        if logger.getEffectiveLevel() == logging.DEBUG:
            logger.debug("SimulationProblem: Found states {}".format(
                ', '.join([var.name() for var in self.__mx['states']])))
            logger.debug("SimulationProblem: Found derivatives {}".format(
                ', '.join([var.name() for var in self.__mx['derivatives']])))
            logger.debug("SimulationProblem: Found algebraics {}".format(
                ', '.join([var.name() for var in self.__mx['algebraics']])))
            logger.debug("SimulationProblem: Found constant inputs {}".format(
                ', '.join([var.name() for var in self.__mx['constant_inputs']])))
            logger.debug("SimulationProblem: Found parameters {}".format(
                ', '.join([var.name() for var in self.__mx['parameters']])))

        # Store the types in an AliasDict
        self.__python_types = AliasDict(self.alias_relation)
        model_variable_types = ["states", "der_states", "alg_states", "inputs", "constants", "parameters"]
        for t in model_variable_types:
            for v in getattr(self.__pymoca_model, t):
                self.__python_types[v.symbol.name()] = v.python_type

        # Store the nominals in an AliasDict
        self.__nominals = AliasDict(self.alias_relation)
        for v in itertools.chain(
                self.__pymoca_model.states, self.__pymoca_model.alg_states):
            sym_name = v.symbol.name()

            # If the nominal is 0.0 or 1.0 or -1.0, ignore: get_variable_nominal returns a default of 1.0
            # TODO: handle nominal vectors (update() will need to load them)
            if ca.MX(v.nominal).is_zero() or ca.MX(v.nominal - 1).is_zero() or ca.MX(v.nominal + 1).is_zero():
                continue
            else:
                if ca.MX(v.nominal).size1() != 1:
                    logger.error('Vector Nominals not supported yet. ({})'.format(sym_name))
                self.__nominals[sym_name] = ca.fabs(v.nominal)
                if logger.getEffectiveLevel() == logging.DEBUG:
                    logger.debug("SimulationProblem: Setting nominal value for variable {} to {}".format(
                        sym_name, self.__nominals[sym_name]))

        # Initialize DAE and initial residuals
        variable_lists = ['states', 'der_states', 'alg_states', 'inputs', 'constants', 'parameters']
        function_arguments = [self.__pymoca_model.time] + [
            ca.veccat(*[v.symbol for v in getattr(self.__pymoca_model, variable_list)])
            for variable_list in variable_lists]

        self.__dae_residual = self.__pymoca_model.dae_residual_function(*function_arguments)

        self.__initial_residual = self.__pymoca_model.initial_residual_function(*function_arguments)
        if self.__initial_residual is None:
            self.__initial_residual = ca.MX()

        # Construct state vector
        self.__sym_list = self.__mx['states'] + self.__mx['algebraics'] + self.__mx['derivatives'] + \
            self.__mx['time'] + self.__mx['constant_inputs'] + self.__mx['parameters']
        self.__state_vector = np.full(len(self.__sym_list), np.nan)

        # A very handy index
        self.__states_end_index = len(self.__mx['states']) + \
            len(self.__mx['algebraics']) + len(self.__mx['derivatives'])

        # NOTE: Backwards compatibility allowing set_var() for parameters. These
        # variables check that this is only done before calling initialize().
        self.__parameters = AliasDict(self.alias_relation)
        self.__parameters.update({v.name(): v for v in self.__mx['parameters']})
        self.__parameters_set_var = True

        # Construct a dict to look up symbols by name (or iterate over)
        self.__sym_dict = OrderedDict(((sym.name(), sym) for sym in self.__sym_list))

        # Generate a dictionary that we can use to lookup the index in the state vector.
        # To avoid repeated and relatively expensive `canonical_signed` calls, we
        # make a dictionary for all variables and their aliases.
        self.__indices = {}
        for i, k in enumerate(self.__sym_dict.keys()):
            for alias in self.alias_relation.aliases(k):
                if alias.startswith('-'):
                    self.__indices[alias[1:]] = (i, -1.0)
                else:
                    self.__indices[alias] = (i, 1.0)

        # Assemble some symbolics, including those needed for a backwards Euler derivative approximation
        X = ca.vertcat(*self.__sym_list[:self.__states_end_index])
        X_prev = ca.vertcat(*[ca.MX.sym(sym.name() + '_prev') for sym in self.__sym_list[:self.__states_end_index]])
        dt = ca.MX.sym("delta_t")

        # Make a list of derivative approximations using backwards Euler formulation
        derivative_approximation_residuals = []
        for index, derivative_state in enumerate(self.__mx['derivatives']):
            derivative_approximation_residuals.append(derivative_state - (X[index] - X_prev[index]) / dt)

        if self.__pymoca_model.delay_states and not self._force_zero_delay:
            raise NotImplementedError("Delayed states are not supported")

        # Delayed feedback (assuming zero delay)
        # TODO: implement delayed feedback support for delay != 0
        delayed_feedback_equations = []
        for delay_state, delay_argument in zip(self.__pymoca_model.delay_states,
                                               self.__pymoca_model.delay_arguments):
            logger.warning("Assuming zero delay for delay state '{}'".format(delay_state))
            delayed_feedback_equations.append(delay_argument.expr - self.__sym_dict[delay_state])

        # Append residuals for derivative approximations
        dae_residual = ca.vertcat(self.__dae_residual, *derivative_approximation_residuals, *delayed_feedback_equations)

        # TODO: implement lookup_tables

        # Make a list of unscaled symbols and a list of their scaled equivalent
        unscaled_symbols = []
        scaled_symbols = []
        for sym_name, nominal in self.__nominals.items():
            # Note that sym_name is always a canonical state
            index, _ = self.__indices[sym_name]

            # If the symbol is a state, Add the symbol to the lists
            if index <= self.__states_end_index:
                unscaled_symbols.append(X[index])
                scaled_symbols.append(X[index] * nominal)

                # Also scale previous states
                unscaled_symbols.append(X_prev[index])
                scaled_symbols.append(X_prev[index] * nominal)

        # Substitute unscaled terms for scaled terms
        dae_residual = ca.substitute(dae_residual, ca.vertcat(*unscaled_symbols), ca.vertcat(*scaled_symbols))

        if logger.getEffectiveLevel() == logging.DEBUG:
            logger.debug('SimulationProblem: DAE Residual is ' + str(dae_residual))

        if X.size1() != dae_residual.size1():
            logger.error('Formulation Error: Number of states ({}) does not equal number of equations ({})'.format(
                X.size1(), dae_residual.size1()))

        # Construct function parameters
        parameters = ca.vertcat(X_prev, *self.__sym_list[self.__states_end_index:])

        # Construct a function res_vals that returns the numerical residuals of a numerical state
        self.__res_vals = ca.Function("res_vals", [X, dt, parameters], [dae_residual])

        # Use rootfinder() to make a function that takes a step forward in time by trying to zero res_vals()
        options = {'nlpsol': 'ipopt', 'nlpsol_options': self.solver_options(), 'error_on_fail': False}
        self.__do_step = ca.rootfinder("next_state", "nlpsol", self.__res_vals, options)

        # Call parent class for default behaviour.
        super().__init__(**kwargs)

    def initialize(self, config_file=None):
        """
        Initialize state vector with default values

        :param config_file: Path to an initialization file.
        """
        if config_file:
            # TODO read start and stop time from config_file and call:
            # self.setup_experiment(start,stop)
            # for now, assume that setup_experiment was called beforehand
            raise NotImplementedError

        # Set values of parameters defined in the model into the state vector
        for var in self.__pymoca_model.parameters:
            # First check to see if parameter is already set (this allows child classes to override model defaults)
            if np.isfinite(self.get_var(var.symbol.name())):
                continue

            # Also test to see if the value is constant
            if isinstance(var.value, ca.MX) and not var.value.is_constant():
                continue

            # Try to extract the value
            try:
                # Extract the value as a python type
                val = var.python_type(var.value)
            except ValueError:
                # var.value is a float NaN being cast to non-float
                continue
            else:
                # If val is finite, we set it
                if np.isfinite(val):
                    logger.debug('SimulationProblem: Setting parameter {} = {}'.format(var.symbol.name(), val))
                    self.set_var(var.symbol.name(), val)

        # Nominals can be symbolic, written in terms of parameters. After all
        # parameter values are known, we evaluate the numeric values of the
        # nominals.
        nominal_vars = list(self.__nominals.keys())
        symbolic_nominals = ca.vertcat(*[self.get_variable_nominal(v) for v in nominal_vars])
        nominal_evaluator = ca.Function('nominal_evaluator', self.__mx['parameters'], [symbolic_nominals])

        n_parameters = len(self.__mx['parameters'])
        if n_parameters > 0:
            [evaluated_nominals] = nominal_evaluator.call(self.__state_vector[-n_parameters:])
        else:
            [evaluated_nominals] = nominal_evaluator.call([])

        evaluated_nominals = np.array(evaluated_nominals).ravel()

        nominal_dict = dict(zip(nominal_vars, evaluated_nominals))

        self.__nominals.update(nominal_dict)

        # Assemble initial residuals and set values from start attributes into the state vector
        minimized_residuals = []
        for var in itertools.chain(self.__pymoca_model.states, self.__pymoca_model.alg_states):
            var_name = var.symbol.name()
            var_nominal = self.get_variable_nominal(var_name)

            # Attempt to cast var.start to python type
            mx_start = ca.MX(var.start)
            if mx_start.is_constant():
                # cast var.start to python type
                start_val = var.python_type(mx_start.to_DM())
            else:
                # var.start is a symbolic expression with unknown value
                start_val = None

            if start_val == 0.0 and not var.fixed:
                # To make initialization easier, we allow setting initial states by providing timeseries
                # with names that match a symbol in the model. We only check for this matching if the start
                # and fixed attributes were left as default
                try:
                    start_val = self.initial_state()[var_name]
                except KeyError:
                    pass
                else:
                    # An initial state was found- add it to the constrained residuals
                    logger.debug('Initialize: Added {} = {} to initial equations (found matching timeseries).'.format(
                        var_name, start_val))
                    # Set var to be fixed
                    var.fixed = True

            if not var.fixed:
                # To make initialization easier, we allow setting initial guesses by providing timeseries
                # with names that match a symbol in the model. We only check for this matching if the start
                # and fixed attributes were left as default
                try:
                    start_val = self.seed()[var_name]
                except KeyError:
                    pass
                else:
                    # An initial state was found- add it to the constrained residuals
                    logger.debug('Initialize: Added {} = {} as initial guess (found matching timeseries).'.format(
                        var_name, start_val))

            # Attempt to set start_val in the state vector. Default to zero if unknown.
            try:
                self.set_var(var_name, start_val if start_val is not None else 0.0)
            except KeyError:
                logger.warning('Initialize: {} not found in state vector. Initial value of {} not set.'.format(
                    var_name, start_val))

            # Add a residual for the difference between the state and its starting expression
            start_expr = start_val if start_val is not None else var.start
            if var.fixed:
                # Set bounds to be equal to each other, such that IPOPT can
                # turn the decision variable into a parameter.
                var.min = start_expr
                var.max = start_expr
            else:
                # minimize residual
                minimized_residuals.append((var.symbol - start_expr) / var_nominal)

        # Default start var for ders is zero
        for der_var in self.__mx['derivatives']:
            self.set_var(der_var.name(), 0.0)

        # Warn for nans in state vector (verify we didn't miss anything)
        self.__warn_for_nans()

        # Optionally encourage a steady-state initial condition
        if getattr(self, 'encourage_steady_state_initial_conditions', False):
            # add penalty for der(var) != 0.0
            for d in self.__mx['derivatives']:
                logger.debug('Added {} to the minimized residuals.'.format(d.name()))
                minimized_residuals.append(d)

        # Make minimized_residuals into a single symbolic object
        minimized_residual = ca.vertcat(*minimized_residuals)

        # Assemble symbolics needed to make a function describing the initial condition of the model
        # We constrain every entry in this MX to zero
        equality_constraints = ca.vertcat(self.__dae_residual, self.__initial_residual)

        # The variables that need a mutually consistent initial condition
        X = ca.vertcat(*self.__sym_list[:self.__states_end_index])

        # Make a list of unscaled symbols and a list of their scaled equivalent
        unscaled_symbols = []
        scaled_symbols = []
        for sym_name, nominal in self.__nominals.items():
            if nominal == 1.0:
                # Nothing to scale
                continue

            # Add the symbol to the lists
            symbol = self.__sym_dict[sym_name]
            unscaled_symbols.append(symbol)
            scaled_symbols.append(symbol * nominal)

        # Make the lists symbolic
        unscaled_symbols = ca.vertcat(*unscaled_symbols)
        scaled_symbols = ca.vertcat(*scaled_symbols)

        # Substitute unscaled terms for scaled terms
        equality_constraints = ca.substitute(equality_constraints, unscaled_symbols, scaled_symbols)
        minimized_residual = ca.substitute(minimized_residual, unscaled_symbols, scaled_symbols)

        logger.debug('SimulationProblem: Initial Equations are ' + str(equality_constraints))
        logger.debug('SimulationProblem: Minimized Residuals are ' + str(minimized_residual))

        # State bounds can be symbolic, written in terms of parameters. After all
        # parameter values are known, we evaluate the numeric values of bounds.
        bound_vars = self.__pymoca_model.states + self.__pymoca_model.alg_states + self.__pymoca_model.der_states
        symbolic_bounds = ca.vertcat(*[ca.horzcat(v.min, v.max) for v in bound_vars])
        bound_evaluator = ca.Function('bound_evaluator', self.__mx['parameters'], [symbolic_bounds])

        # Evaluate bounds using values of parameters
        n_parameters = len(self.__mx['parameters'])
        if n_parameters > 0:
            [evaluated_bounds] = bound_evaluator.call(self.__state_vector[-n_parameters:])
        else:
            [evaluated_bounds] = bound_evaluator.call([])

        # Scale the bounds with the nominals
        nominals = []
        for var in bound_vars:
            nominals.append(self.get_variable_nominal(var.symbol.name()))

        evaluated_bounds = np.array(evaluated_bounds) / np.array(nominals)[:, None]

        # Update with the bounds of delayed states
        n_delay = len(self.__pymoca_model.delay_states)
        delay_bounds = np.array([-np.inf, np.inf] * n_delay).reshape((n_delay, 2))
        offset = len(self.__pymoca_model.states) + len(self.__pymoca_model.alg_states)
        evaluated_bounds = np.vstack((evaluated_bounds[:offset, :],
                                      delay_bounds,
                                      evaluated_bounds[offset:, :]))

        # Construct arrays of state bounds (used in the initialize() nlp, but not in __do_step rootfinder)
        self.__lbx = evaluated_bounds[:, 0]
        self.__ubx = evaluated_bounds[:, 1]

        # Constrain model equation residuals to zero
        lbg = np.zeros(equality_constraints.size1())
        ubg = np.zeros(equality_constraints.size1())

        # Construct objective function from the input residual
        objective_function = ca.dot(minimized_residual, minimized_residual)

        # Construct nlp and solver to find initial state using ipopt
        parameters = ca.vertcat(*self.__mx['time'], *self.__mx['constant_inputs'], *self.__mx['parameters'])
        nlp = {'x': X, 'f': objective_function, 'g': equality_constraints, 'p': parameters}
        solver = ca.nlpsol('solver', 'ipopt', nlp, self.solver_options())

        # Construct guess
        guess = ca.vertcat(*np.nan_to_num(self.__state_vector[:self.__states_end_index]))

        # Find initial state
        initial_state = solver(x0=guess,
                               lbx=self.__lbx, ubx=self.__ubx,
                               lbg=lbg, ubg=ubg,
                               p=self.__state_vector[self.__states_end_index:])

        # If unsuccessful, stop.
        return_status = solver.stats()['return_status']
        if return_status not in {'Solve_Succeeded', 'Solved_To_Acceptable_Level'}:
            raise Exception('Initialization Failed with return status "{}"'.format(return_status))

        # Update state vector with initial conditions
        self.__state_vector[:self.__states_end_index] = initial_state['x'][:self.__states_end_index].T

        # make a copy of the initialized initial state vector in case we want to run the model again
        self.__initialized_state_vector = copy.deepcopy(self.__state_vector)

        # Warn for nans in state vector after initialization
        self.__warn_for_nans()

        # No longer allow setting parameters with set_var(), as we want to be
        # clear that that does not work
        self.__parameters_set_var = False

        self.__parameter_names_including_aliases = set()
        for p in self.__parameters.keys():
            self.__parameter_names_including_aliases |= self.alias_relation.aliases(p)

    def pre(self):
        """
        Any preprocessing takes place here.
        """
        pass

    def post(self):
        """
        Any postprocessing takes place here.
        """
        pass

    def setup_experiment(self, start, stop, dt):
        """
        Method for subclasses (PIMixin, CSVMixin, or user classes) to set timing information for a simulation run.

        :param start: Start time for the simulation.
        :param stop:  Final time for the simulation.
        :param dt:    Time step size.
        """

        # Set class vars with start/stop/dt values
        self.__start = start
        self.__stop = stop
        self.__dt = dt

        # Set time in state vector
        self.set_var('time', start)

    def update(self, dt):
        """
        Performs one timestep.

        The methods ``setup_experiment`` and ``initialize`` must have been called before.

        :param dt: Time step size.
        """
        if dt < 0:
            dt = self.__dt

        logger.debug("Taking a step at {} with size {}".format(self.get_current_time(), dt))

        # increment time
        self.set_var('time', self.get_current_time() + dt)

        # take a step
        guess = self.__state_vector[:self.__states_end_index]
        next_state = self.__do_step(guess, dt, self.__state_vector)

        # Check convergence of rootfinder
        rootfinder_stats = self.__do_step.stats()

        if not rootfinder_stats['success']:
            logger.warning(
                'Simulation has failed to converge at time {}. Solver failed with status {}'.format(
                    self.get_current_time(), rootfinder_stats['nlpsol']['return_status']))

        if logger.getEffectiveLevel() == logging.DEBUG:
            # compute max residual
            largest_res = ca.norm_inf(self.__res_vals(next_state, self.__dt, self.__state_vector))
            logger.debug('Residual maximum magnitude: {:.2E}'.format(float(largest_res)))

        # Update state vector
        self.__state_vector[:self.__states_end_index] = next_state.toarray().ravel()

    def simulate(self):
        """
        Run model from start_time to end_time.
        """

        # Do any preprocessing, which may include changing parameter values on
        # the model
        logger.info("Preprocessing")
        self.pre()

        # Initialize model
        logger.info("Initializing")
        self.initialize()

        # Perform all timesteps
        logger.info("Running")
        while self.get_current_time() < self.get_end_time():
            self.update(-1)

        # Do any postprocessing
        logger.info("Postprocessing")
        self.post()

    def reset(self):
        """
        Reset the FMU.
        """
        self.__state_vector = copy.deepcopy(self.__initialized_state_vector)

    def get_start_time(self):
        """
        Return start time of experiment.

        :returns: The start time of the experiment.
        """
        return self.__start

    def get_end_time(self):
        """
        Return end time of experiment.

        :returns: The end time of the experiment.
        """
        return self.__stop

    def get_current_time(self):
        """
        Return current time of simulation.

        :returns: The current simulation time.
        """
        return self.get_var('time')

    def get_time_step(self):
        """
        Return simulation timestep.

        :returns: The simulation timestep.
        """
        return self.__dt

    def get_var(self, name):
        """
        Return a numpy array from FMU.

        :param name: Variable name.

        :returns: The value of the variable.
        """

        # Get the index of the canonical state and sign
        index, sign = self.__indices[name]
        value = self.__state_vector[index]

        # Adjust sign if needed
        if sign < 0:
            value *= sign

        # Adjust for nominal value if not default
        if index <= self.__states_end_index:
            nominal = self.get_variable_nominal(name)
            value *= nominal

        return value

    def get_var_count(self):
        """
        Return the number of variables in the model.

        :returns: The number of variables in the model.
        """
        return len(self.get_variables())

    def get_var_name(self, i):
        """
        Returns the name of a variable.

        :param i: Index in ordered dictionary returned by method get_variables.

        :returns: The name of the variable.
        """
        return list(self.get_variables())[i]

    def get_var_type(self, name):
        """
        Return type, compatible with numpy.

        :param name: String variable name.

        :returns: The numpy-compatible type of the variable.

        :raises: KeyError
        """
        return self.__python_types[name]

    def get_var_rank(self, name):
        """
        Not implemented
        """
        raise NotImplementedError

    def get_var_shape(self, name):
        """
        Not implemented
        """
        raise NotImplementedError

    def get_variables(self):
        """
        Return all variables (both internal and user defined)

        :returns: An ordered dictionary of all variables supported by the model.
        """
        return self.__sym_dict

    @cached
    def get_state_variables(self):
        return AliasDict(
            self.alias_relation,
            {sym.name(): sym for sym in (self.__mx['states'] + self.__mx['algebraics'])})

    @cached
    def get_parameter_variables(self):
        return AliasDict(
            self.alias_relation,
            {sym.name(): sym for sym in self.__mx['parameters']})

    @cached
    def get_input_variables(self):
        return AliasDict(
            self.alias_relation,
            {sym.name(): sym for sym in self.__mx['constant_inputs']})

    @cached
    def get_output_variables(self):
        return self.__pymoca_model.outputs

    def __warn_for_nans(self):
        """
        Test state vector for missing values and warn
        """
        value_is_nan = np.isnan(self.__state_vector)
        if any(value_is_nan):
            for sym, isnan in zip(self.__sym_list, value_is_nan):
                if isnan:
                    logger.warning('Variable {} has no value.'.format(sym))

    def set_var(self, name, value):
        """
        Set the value of the given variable.

        :param name: Name of variable to set.
        :param value:  Value(s).
        """

        # TODO: sanitize input

        # Check if it is a parameter, and if it is allowed to be set
        if not self.__parameters_set_var:
            if name in self.__parameter_names_including_aliases:
                raise Exception("Cannot set parameters after initialize() has been called.")

        # Get the index of the canonical state and sign
        index, sign = self.__indices[name]
        if sign < 0:
            value *= sign

        # Adjust for nominal value if not default
        if index <= self.__states_end_index:
            nominal = self.get_variable_nominal(name)
            value /= nominal

        # Store value in state vector
        self.__state_vector[index] = value

    def set_var_slice(self, name, start, count, var):
        """
        Not implemented.
        """
        raise NotImplementedError

    def set_var_index(self, name, index, var):
        """
        Not implemented.
        """
        raise NotImplementedError

    def inq_compound(self, name):
        """
        Not implemented.
        """
        raise NotImplementedError

    def inq_compound_field(self, name, index):
        """
        Not implemented.
        """
        raise NotImplementedError

    def solver_options(self):
        """
        Returns a dictionary of CasADi root_finder() solver options.

        :returns: A dictionary of CasADi :class:`root_finder` options.  See the CasADi documentation for details.
        """
        return {'ipopt.fixed_variable_treatment': 'make_parameter',
                'ipopt.print_level': 0,
                'print_time': False,
                'error_on_fail': False}

    def get_variable_nominal(self, variable) -> Union[float, ca.MX]:
        """
        Get the value of the nominal attribute of a variable

        NOTE: Due to backwards compatibility for allowing parameters to be set
        with set_var() instead of overriding parameters(), this method can
        return a symbolic value for nominals defined in the Modelica file. It
        can only do so until the initializion() method in this class is
        called/completed, after which it will return numeric values only.
        """
        return self.__nominals.get(variable, 1.0)

    def timeseries_at(self, variable, t):
        """
        Get value of timeseries variable at time t: should be overridden by pi or csv mixin
        """
        raise NotImplementedError

    @cached
    def initial_state(self) -> AliasDict[str, float]:
        """
        The initial state.

        :returns: A dictionary of variable names and initial state (t0) values.
        """
        t0 = self.get_start_time()
        initial_state_dict = AliasDict(self.alias_relation)

        for variable in list(self.get_state_variables()) + list(self.get_input_variables()):
            try:
                initial_state_dict[variable] = self.timeseries_at(variable, t0)
            except KeyError:
                pass
            except NotImplementedError:
                pass
            else:
                if logger.getEffectiveLevel() == logging.DEBUG:
                    logger.debug("Read intial state for {}".format(variable))

        return initial_state_dict

    @cached
    def seed(self) -> AliasDict[str, float]:
        """
        Seed values providing an initial guess for the t0 states.

        :returns: A dictionary of variable names and seed (t0) values.
        """
        return AliasDict(self.alias_relation)

    @cached
    def parameters(self):
        """
        Return a dictionary of parameter values extracted from the Modelica model
        """
        # Create AliasDict
        parameters = AliasDict(self.alias_relation)

        # Update with model parameters
        parameters.update({p.symbol.name(): p.value for p in self.__pymoca_model.parameters})

        return parameters

    @property
    @cached
    def alias_relation(self):
        return self.__pymoca_model.alias_relation

    @cached
    def compiler_options(self):
        """
        Subclasses can configure the `pymoca <http://github.com/pymoca/pymoca>`_ compiler options here.

        :returns: A dictionary of pymoca compiler options.  See the pymoca documentation for details.
        """

        # Default options
        compiler_options = {}

        # Expand vector states to multiple scalar component states.
        compiler_options['expand_vectors'] = True

        # Where imported model libraries are located.
        library_folders = self.modelica_library_folders.copy()

        for ep in pkg_resources.iter_entry_points(group='rtctools.libraries.modelica'):
            if ep.name == "library_folder":
                library_folders.append(
                    pkg_resources.resource_filename(ep.module_name, ep.attrs[0]))

        compiler_options['library_folders'] = library_folders

        # Eliminate equations of the type 'var = const'.
        compiler_options['eliminate_constant_assignments'] = True

        # Eliminate constant symbols from model, replacing them with the values
        # specified in the model.
        compiler_options['replace_constant_values'] = True

        # Replace any constant expressions into the model.
        compiler_options['replace_constant_expressions'] = True

        # Replace any parameter expressions into the model.
        compiler_options['replace_parameter_expressions'] = True

        # Eliminate variables starting with underscores.
        compiler_options['eliminable_variable_expression'] = r'(.*[.]|^)_\w+(\[[\d,]+\])?\Z'

        # Pymoca currently requires `expand_mx` to be set for
        # `eliminable_variable_expression` to work.
        compiler_options['expand_mx'] = True

        # Automatically detect and eliminate alias variables.
        compiler_options['detect_aliases'] = True

        # Cache the model on disk
        compiler_options['cache'] = True

        # Done
        return compiler_options



import logging
from collections import OrderedDict
from typing import Tuple

import rtctools.data.netcdf as netcdf
from rtctools.optimization.io_mixin import IOMixin

logger = logging.getLogger("rtctools")


class NetCDFMixin(IOMixin):
    """
    Adds NetCDF I/O to your optimization problem.

    During preprocessing, a file named timeseries_import.nc is read from the ``input`` subfolder.
    During postprocessing a file named timeseries_export.nc is written to the ``output`` subfolder.

    Both the input and output nc files are expected to follow the FEWS format for
    scalar data in a NetCDF file, i.e.:

    - They must contain a variable with the station ids (location ids) which can
      be recognized by the attribute `cf_role` set to `timeseries_id`.
    - They must contain a time variable with attributes `standard_name` = `time`
      and `axis` = `T`

    From the input file, all 2-D (or 3-D in case of ensembles) variables with dimensions equal
    to the station ids and time variable (and realization) are read.

    To map the NetCDF parameter identifier to and from an RTC-Tools variable name,
    the overridable methods :py:meth:`netcdf_id_to_variable` and
    :py:meth:`netcdf_id_from_variable` are used.

    :cvar netcdf_validate_timeseries:
        Check consistency of timeseries. Default is ``True``
    """

    #: Check consistency of timeseries.
    netcdf_validate_timeseries = True

    def netcdf_id_to_variable(self, station_id: str, parameter: str) -> str:
        """
        Maps the station_id and the parameter name to the variable name to be
        used in RTC-Tools.

        :return: The variable name used in RTC-Tools
        """
        return '{}__{}'.format(station_id, parameter)

    def netcdf_id_from_variable(self, variable_name: str) -> Tuple[str, str]:
        """
        Maps the variable name in RTC-Tools to a station_id and parameter name
        for writing to a NetCDF file.

        :return: A pair of station_id and parameter
        """
        return variable_name.split("__")

    def read(self):
        # Call parent class first for default behaviour
        super().read()

        dataset = netcdf.ImportDataset(self._input_folder, self.timeseries_import_basename)
        # Although they are not used outside of this method, we add some
        # variables to self for debugging purposes
        self.__timeseries_import = dataset

        # store the import times
        times = self.__timeseries_times = dataset.read_import_times()
        self.io.reference_datetime = self.__timeseries_times[0]

        # Timestamp check
        self.__dt = times[1] - times[0] if len(times) >= 2 else 0
        for i in range(len(times) - 1):
            if times[i + 1] - times[i] != self.__dt:
                self.__dt = None
                break

        if self.netcdf_validate_timeseries:
            # check if strictly increasing
            for i in range(len(times) - 1):
                if times[i] >= times[i + 1]:
                    raise Exception('NetCDFMixin: Time stamps must be strictly increasing.')

        # store the station data for later use
        self.__stations = dataset.read_station_data()
        # read all available timeseries from the dataset
        timeseries_var_keys = dataset.find_timeseries_variables()

        for parameter in timeseries_var_keys:
            for i, station_id in enumerate(self.__stations.station_ids):
                name = self.netcdf_id_to_variable(station_id, parameter)

                if dataset.ensemble_member_variable is not None:
                    if dataset.ensemble_member_variable.dimensions[0] in dataset.variable_dimensions(parameter):
                        for ensemble_member_index in range(self.__timeseries_import.ensemble_size):
                            values = dataset.read_timeseries_values(i, parameter, ensemble_member_index)
                            self.io.set_timeseries(name, self.__timeseries_times, values, ensemble_member_index)
                    else:
                        values = dataset.read_timeseries_values(i, parameter, 0)
                        for ensemble_member_index in range(self.__timeseries_import.ensemble_size):
                            self.io.set_timeseries(name, self.__timeseries_times, values, ensemble_member_index)
                else:
                    values = dataset.read_timeseries_values(i, parameter, 0)
                    self.io.set_timeseries(name, self.__timeseries_times, values, 0)

                logger.debug('Read timeseries data for station id "{}" and parameter "{}", '
                             'stored under variable name "{}"'
                             .format(station_id, parameter, name))

        logger.debug("NetCDFMixin: Read timeseries")

    def write(self):
        # Call parent class first for default behaviour
        super().write()

        dataset = netcdf.ExportDataset(self._output_folder, self.timeseries_export_basename)

        times = [(dt - self.__timeseries_times[0]).seconds for dt in self.__timeseries_times]
        dataset.write_times(times, self.initial_time, self.io.reference_datetime)

        output_variables = [sym.name() for sym in self.output_variables]

        output_station_ids, output_parameter_ids = zip(*(
            self.netcdf_id_from_variable(var_name) for var_name in output_variables))

        # Make sure that output_station_ids and output_parameter_ids are
        # unique, but make sure to avoid non-deterministic ordering.
        unique_station_ids = list(OrderedDict.fromkeys(output_station_ids))
        unique_parameter_ids = list(OrderedDict.fromkeys(output_parameter_ids))

        dataset.write_station_data(self.__stations, unique_station_ids)
        dataset.write_ensemble_data(self.ensemble_size)

        dataset.create_variables(unique_parameter_ids, self.ensemble_size)

        for ensemble_member in range(self.ensemble_size):
            results = self.extract_results(ensemble_member)

            for var_name, station_id, parameter_id in zip(
                    output_variables, output_station_ids, output_parameter_ids):
                # determine the output values
                try:
                    values = results[var_name]
                    if len(values) != len(times):
                        values = self.interpolate(
                            times, self.times(var_name), values, self.interpolation_method(var_name))
                except KeyError:
                    try:
                        ts = self.get_timeseries(var_name, ensemble_member)
                        if len(ts.times) != len(times):
                            values = self.interpolate(
                                times, ts.times, ts.values)
                        else:
                            values = ts.values
                    except KeyError:
                        logger.error(
                            'NetCDFMixin: Output requested for non-existent variable {}. '
                            'Will not be in output file.'.format(var_name))
                        continue

                dataset.write_output_values(station_id, parameter_id, ensemble_member, values, self.ensemble_size)

        dataset.close()



import itertools
import logging
from collections import OrderedDict
from enum import Enum
from typing import Dict, Union

import casadi as ca

import numpy as np

from .goal_programming_mixin import GoalProgrammingMixin
from .goal_programming_mixin_base import Goal, StateGoal, _GoalProgrammingMixinBase  # noqa: F401
from .goal_programming_mixin_base import _EmptyEnsembleList, _EmptyEnsembleOrderedDict
from .timeseries import Timeseries

logger = logging.getLogger("rtctools")


class SinglePassMethod(Enum):
    APPEND_CONSTRAINTS_OBJECTIVE = 1
    UPDATE_OBJECTIVE_CONSTRAINT_BOUNDS = 2


class SinglePassGoalProgrammingMixin(_GoalProgrammingMixinBase):
    r"""
    Adds lexicographic goal programming to your optimization problem.

    Unlike :py:class:`.GoalProgrammingMixin`, this mixin will call
    :py:meth:`.transcribe` only once per call to :py:meth:`.optimize`, and not
    :math:`N` times for :math:`N` priorities. It works similar to how
    `keep_soft_constraints = True` works for :py:class:`.GoalProgrammingMixin`,
    while avoiding the repeated calls to transcribe the problem.

    This mixin can work in one of two ways. What is shared between them is
    that all violation variables of all goals are generated once at the
    beginning, such that the state vector is exactly the same for all
    priorities. They also share that all goal constraints are added from the
    start. How they differ is in how they handle/append the constraints on the
    objective of previous priorities:

    1. At priority :math:`i` the constraints are the same as the ones at
       priority :math:`i - 1` with the addition of the objective constraint
       related to priority :math:`i - 1`. This is the default method.

    2. All objective constraints are added at the start. The objective
       constraints will have bound of :math:`[-\inf, \inf]` at the start, to be
       updated after each priority finishes.

    There is a special `qpsol` alternative available :py:class:`CachingQPSol`,
    that will avoid recalculations on constraints that were already there in
    previous priorities. This works for both options outlined above, because
    the assumptions of :py:class:`CachingQPSol` are that:

    1. The state vector does not change
    2. Any new constraints are appended at the end

    .. note::

        Just like GoalProgrammingMixin, objective constraints are only added on
        the goal objectives, not on any custom user objective.
    """

    single_pass_method = SinglePassMethod.APPEND_CONSTRAINTS_OBJECTIVE

    def __init__(self, **kwargs):
        # Call parent class first for default behaviour.
        super().__init__(**kwargs)

        # Initialize instance variables, so that the overridden methods may be
        # called outside of the goal programming loop, for example in pre().
        self._gp_first_run = True
        self.__results_are_current = False

        self.__constraint_store = _EmptyEnsembleOrderedDict()
        self.__path_constraint_store = _EmptyEnsembleOrderedDict()

        self.__problem_constraints = _EmptyEnsembleList()
        self.__problem_path_constraints = _EmptyEnsembleList()
        self.__problem_epsilons = []
        self.__problem_path_epsilons = []
        self.__problem_path_timeseries = []
        self.__problem_parameters = []

        self.__current_priority = 0
        self.__original_constraints = None
        self.__previous_constraints = None

        self.__soft_constraints_per_priority = []
        self.__path_soft_constraints_per_priority = []

        self.__objectives_per_priority = []
        self.__path_objectives_per_priority = []

        if isinstance(self, GoalProgrammingMixin):
            raise Exception("Cannot be an instance of both GoalProgrammingMixin and SinglePassGoalProgrammingMixin")

    @property
    def extra_variables(self):
        return self.__problem_epsilons

    @property
    def path_variables(self):
        return self.__problem_path_epsilons

    def bounds(self):
        bounds = super().bounds()
        for epsilon in (self.__problem_epsilons + self.__problem_path_epsilons):
            bounds[epsilon.name()] = (0.0, 1.0)
        return bounds

    def constant_inputs(self, ensemble_member):
        constant_inputs = super().constant_inputs(ensemble_member)

        n_times = len(self.times())

        # Append min/max timeseries to the constant inputs. Note that min/max
        # timeseries are shared between all ensemble members.
        for (variable, value) in self.__problem_path_timeseries:
            if isinstance(value, np.ndarray):
                value = Timeseries(self.times(), np.broadcast_to(value, (n_times, len(value))))
            elif not isinstance(value, Timeseries):
                value = Timeseries(self.times(), np.full(n_times, value))

            constant_inputs[variable] = value
        return constant_inputs

    def parameters(self, ensemble_member):
        parameters = super().parameters(ensemble_member)

        # Append min/max values to the parameters. Note that min/max values
        # are shared between all ensemble members.
        for (variable, value) in self.__problem_parameters:
            parameters[variable] = value

        return parameters

    def seed(self, ensemble_member):
        assert self._gp_first_run
        return super().seed(ensemble_member)

    def constraints(self, ensemble_member):
        constraints = super().constraints(ensemble_member)

        additional_constraints = itertools.chain(
            self.__constraint_store[ensemble_member].values(),
            self.__problem_constraints[ensemble_member])

        for constraint in additional_constraints:
            constraints.append((constraint.function(self), constraint.min, constraint.max))

        return constraints

    def path_constraints(self, ensemble_member):
        path_constraints = super().path_constraints(ensemble_member)

        additional_path_constraints = itertools.chain(
            self.__path_constraint_store[ensemble_member].values(),
            self.__problem_path_constraints[ensemble_member])

        for constraint in additional_path_constraints:
            path_constraints.append((constraint.function(self), constraint.min, constraint.max))

        return path_constraints

    def solver_options(self):

        # TODO: Split off into private

        # Call parent
        options = super().solver_options()

        solver = options['solver']
        assert solver in ['bonmin', 'ipopt']

        # Make sure constant states, such as min/max timeseries for violation variables,
        # are turned into parameters for the final optimization problem.
        ipopt_options = options[solver]
        ipopt_options['fixed_variable_treatment'] = 'make_parameter'

        # Define temporary variable to avoid infinite loop between
        # solver_options and goal_programming_options.
        self._loop_breaker_solver_options = True

        if not hasattr(self, '_loop_breaker_goal_programming_options'):
            if not self.goal_programming_options()['mu_reinit']:
                ipopt_options['mu_strategy'] = 'monotone'
                if not self._gp_first_run:
                    ipopt_options['mu_init'] = self.solver_stats['iterations'][
                        'mu'][-1]

        delattr(self, '_loop_breaker_solver_options')

        return options

    def goal_programming_options(self) -> Dict[str, Union[float, bool]]:
        """
        Returns a dictionary of options controlling the goal programming process.

        +---------------------------+-----------+---------------+
        | Option                    | Type      | Default value |
        +===========================+===========+===============+
        | ``constraint_relaxation`` | ``float`` | ``0.0``       |
        +---------------------------+-----------+---------------+
        | ``mu_reinit``             | ``bool``  | ``True``      |
        +---------------------------+-----------+---------------+
        | ``fix_minimized_values``  | ``bool``  | ``True/False``|
        +---------------------------+-----------+---------------+
        | ``check_monotonicity``    | ``bool``  | ``True``      |
        +---------------------------+-----------+---------------+
        | ``equality_threshold``    | ``float`` | ``1e-8``      |
        +---------------------------+-----------+---------------+
        | ``scale_by_problem_size`` | ``bool``  | ``False``     |
        +---------------------------+-----------+---------------+

        When a priority's objective is turned into a hard constraint,
        the constraint is relaxed with ``constraint_relaxation``. Use of this option is
        normally not required. Note that:

        When using the default solver (IPOPT), its barrier parameter ``mu`` is
        normally re-initialized at every iteration of the goal programming
        algorithm, unless mu_reinit is set to ``False``.  Use of this option
        is normally not required.

        If ``fix_minimized_values`` is set to ``True``, goal functions will be set to equal their
        optimized values in optimization problems generated during subsequent priorities. Otherwise,
        only an upper bound will be set. Use of this option is normally not required.
        Note that the use of this option may add non-convex constraints to the optimization problem.
        The default value for this parameter is ``True`` for the default solvers IPOPT/BONMIN. If any
        other solver is used, the default value is ``False``.

        If ``check_monotonicity`` is set to ``True``, then it will be checked whether goals with the same
        function key form a monotonically decreasing sequence with regards to the target interval.

        The option ``equality_threshold`` controls when a two-sided inequality constraint is folded into
        an equality constraint.

        If ``scale_by_problem_size`` is set to ``True``, the objective (i.e. the sum of the violation variables)
        will be divided by the number of goals, and the path objective will be divided by the number
        of path goals and the number of active time steps (per goal). This will make sure the objectives are always in
        the range [0, 1], at the cost of solving each goal/time step less accurately.

        :returns: A dictionary of goal programming options.
        """

        options = {}

        options['mu_reinit'] = True
        options['constraint_relaxation'] = 0.0  # Disable by default
        options['fix_minimized_values'] = False
        options['check_monotonicity'] = True
        options['equality_threshold'] = 1e-8
        options['scale_by_problem_size'] = False

        # Forced options to be able to re-use GoalProgrammingMixin's
        # GoalProgrammingMixin._gp_* functions. These are not relevant for
        # SinglePassGoalProgrammingMixin, or should be set to a certain value
        # for it to make sense.
        options['violation_relaxation'] = 0.0  # Disable by default
        options['violation_tolerance'] = np.inf  # Disable by default
        options['interior_distance'] = 1e-6
        options['keep_soft_constraints'] = True

        # Define temporary variable to avoid infinite loop between
        # solver_options and goal_programming_options.
        self._loop_breaker_goal_programming_options = True

        if not hasattr(self, '_loop_breaker_solver_options'):
            if self.solver_options()['solver'] in {'ipopt', 'bonmin'}:
                options['fix_minimized_values'] = True

        delattr(self, '_loop_breaker_goal_programming_options')

        return options

    def optimize(self, preprocessing=True, postprocessing=True, log_solver_failure_as_error=True):
        # Do pre-processing
        if preprocessing:
            self.pre()

        # Group goals into subproblems
        subproblems = []
        goals = self.goals()
        path_goals = self.path_goals()

        options = self.goal_programming_options()

        # Validate goal definitions
        self._gp_validate_goals(goals, is_path_goal=False)
        self._gp_validate_goals(path_goals, is_path_goal=True)

        priorities = sorted({int(goal.priority) for goal in itertools.chain(goals, path_goals) if not goal.is_empty})

        for priority in priorities:
            subproblems.append((
                priority,
                [goal for goal in goals if int(goal.priority) == priority and not goal.is_empty],
                [goal for goal in path_goals if int(goal.priority) == priority and not goal.is_empty]))

        # Solve the subproblems one by one
        logger.info("Starting goal programming")

        success = False

        self.__constraint_store = [OrderedDict() for ensemble_member in range(self.ensemble_size)]
        self.__path_constraint_store = [OrderedDict() for ensemble_member in range(self.ensemble_size)]

        self.__problem_constraints = [[] for ensemble_member in range(self.ensemble_size)]
        self.__problem_path_constraints = [[] for ensemble_member in range(self.ensemble_size)]

        self.__problem_epsilons = []
        self.__problem_parameters = []
        self.__problem_path_epsilons = []
        self.__problem_path_timeseries = []

        self._gp_first_run = True
        self.__results_are_current = False

        self.__current_priority = 0
        self.__original_constraints = None

        self.__objectives_per_priority = []
        self.__path_objectives_per_priority = []

        self.__additional_constraints = []
        self.__objectives = []

        for i, (_, goals, path_goals) in enumerate(subproblems):
            (subproblem_epsilons, subproblem_objectives,
             subproblem_soft_constraints, hard_constraints,
             subproblem_parameters) = \
                self._gp_goal_constraints(goals, i, options, is_path_goal=False)

            (subproblem_path_epsilons, subproblem_path_objectives,
             subproblem_path_soft_constraints, path_hard_constraints,
             subproblem_path_timeseries) = \
                self._gp_goal_constraints(path_goals, i, options, is_path_goal=True)

            # Put hard constraints in the constraint stores
            self._gp_update_constraint_store(self.__constraint_store, hard_constraints)
            self._gp_update_constraint_store(self.__path_constraint_store, path_hard_constraints)

            # Append new variables, parameters, timeseries and constraints to
            # their respective lists
            self.__problem_epsilons.extend(subproblem_epsilons)
            self.__problem_path_epsilons.extend(subproblem_path_epsilons)

            self.__problem_parameters.extend(subproblem_parameters)
            self.__problem_path_timeseries.extend(subproblem_path_timeseries)

            for ensemble_member in range(self.ensemble_size):
                self.__problem_constraints[ensemble_member].extend(
                    subproblem_soft_constraints[ensemble_member])
                self.__problem_path_constraints[ensemble_member].extend(
                    subproblem_path_soft_constraints[ensemble_member])

            self.__objectives_per_priority.append(subproblem_objectives)
            self.__path_objectives_per_priority.append(subproblem_path_objectives)

        for priority in priorities:
            logger.info("Solving goals at priority {}".format(priority))

            # Call the pre priority hook
            self.priority_started(priority)

            # Solve subproblem
            success = super().optimize(
                preprocessing=False, postprocessing=False, log_solver_failure_as_error=log_solver_failure_as_error)
            if not success:
                break

            self._gp_first_run = False

            # To match GoalProgrammingMixin's behavior of applying the
            # constraint_relaxation value at priority 2 on the objective of
            # priority 2 (and not that of priority 1), we have to store the
            # two relevant options here for later use.
            options = self.goal_programming_options()
            self.__objective_constraint_options = {
                k: v for k, v in options.items()
                if k in {'fix_minimized_values', 'constraint_relaxation'}
            }

            # Store results.  Do this here, to make sure we have results even
            # if a subsequent priority fails.
            self.__results_are_current = False
            self.__results = [self.extract_results(
                ensemble_member) for ensemble_member in range(self.ensemble_size)]
            self.__results_are_current = True

            # Call the post priority hook, so that intermediate results can be
            # logged/inspected.
            self.priority_completed(priority)

            self.__current_priority += 1

        logger.info("Done goal programming")

        # Do post-processing
        if postprocessing:
            self.post()

        # Done
        return success

    def transcribe(self):

        def _objective_func(subproblem_objectives, subproblem_path_objectives):
            val = 0.0
            for ensemble_member in range(self.ensemble_size):
                n_objectives = self._gp_n_objectives(
                    subproblem_objectives, subproblem_path_objectives, ensemble_member)
                expr = self._gp_objective(subproblem_objectives, n_objectives, ensemble_member)
                expr += ca.sum1(self.map_path_expression(
                    self._gp_path_objective(subproblem_path_objectives, n_objectives, ensemble_member),
                    ensemble_member))
                val += self.ensemble_member_probability(ensemble_member) * expr

            return val

        if self._gp_first_run:
            discrete, lbx, ubx, lbg, ubg, x0, nlp = super().transcribe()
            self.__original_transcribe = (discrete, lbx, ubx, lbg, ubg, x0, nlp)

            self.__additional_constraints = []
            self.__objectives = []

            # Objectives
            for subproblem_objectives, subproblem_path_objectives in zip(
                    self.__objectives_per_priority, self.__path_objectives_per_priority):
                self.__objectives.append(_objective_func(subproblem_objectives, subproblem_path_objectives))

            if self.single_pass_method == SinglePassMethod.UPDATE_OBJECTIVE_CONSTRAINT_BOUNDS:
                # The objectives are also directly added as constraints
                constraints = [(objective, -np.inf, np.inf) for objective in self.__objectives]
                self.__additional_constraints.extend(constraints)

        # Add constraint on the objective of previous priority
        if self.__current_priority > 0:
            options = self.__objective_constraint_options

            previous_objective = self.__objectives[self.__current_priority - 1]
            f = ca.Function('tmp', [self.solver_input], [previous_objective])
            obj_val = float(f(self.solver_output))

            if options['fix_minimized_values']:
                lb, ub = obj_val, obj_val
                self.linear_collocation = False  # Disable solver option jac_c_constant for IPOPT
            else:
                obj_val += options['constraint_relaxation']
                lb, ub = -np.inf, obj_val

            if self.single_pass_method == SinglePassMethod.APPEND_CONSTRAINTS_OBJECTIVE:
                self.__additional_constraints.append((self.__objectives[self.__current_priority - 1], lb, ub))
            elif self.single_pass_method == SinglePassMethod.UPDATE_OBJECTIVE_CONSTRAINT_BOUNDS:
                ind = self.__current_priority - 1
                constraint = self.__additional_constraints[ind]
                self.__additional_constraints[ind] = (constraint[0], lb, ub)

        # Update the NLP
        discrete, lbx, ubx, lbg, ubg, x0, nlp = self.__original_transcribe
        nlp = nlp.copy()

        if self.__additional_constraints:
            g_extra, lbg_extra, ubg_extra = zip(*self.__additional_constraints)

            g = ca.vertcat(nlp['g'], *g_extra)
            lbg = [*lbg.copy(), *lbg_extra]
            ubg = [*ubg.copy(), *ubg_extra]

            nlp['g'] = g

        nlp['f'] = self.__objectives[self.__current_priority]

        if not self._gp_first_run:
            x0 = self.solver_output.copy()

        return discrete, lbx, ubx, lbg, ubg, x0, nlp

    def extract_results(self, ensemble_member=0):
        if self.__results_are_current:
            logger.debug("Returning cached results")
            return self.__results[ensemble_member]

        # If self.__results is not up to date, do the super().extract_results
        # method
        return super().extract_results(ensemble_member)


class CachingQPSol:
    """
    Alternative to :py:func:`ca.qpsol` that caches the Jacobian between calls.

    Typical usage would be something like:

    .. code-block::

        def pre(self):
            self._qpsol = CachingQPSol()
            super().pre()

        def solver_options():
            options = super().solver_options()
            options['casadi_solver'] = self._qpsol
            return options
    """

    def __init__(self):
        self._tlcache = {}

    def __call__(self, name, solver_name, nlp, options):

        class Solver:
            def __init__(self, nlp=nlp, solver_name=solver_name, options=options, cache=self._tlcache):
                x = nlp['x']
                f = nlp['f']
                g = nlp['g']

                if isinstance(x, ca.MX):
                    # Can only convert SX to DM
                    x = ca.SX.sym('X', *x.shape)
                    x_mx = nlp['x']
                    expand = True
                else:
                    x_mx = None
                    expand = False

                if expand:
                    expand_f = ca.Function('f', [x_mx], [f]).expand()
                    f = expand_f(x)

                # Gradient of the objective: gf == Hx + g
                gf = ca.gradient(f, x)

                # Identify the linear term in the objective
                c = ca.substitute(gf, x, ca.DM.zeros(x.sparsity()))

                # Identify the quadratic term in the objective
                H = 0.5*ca.jacobian(gf, x, {"symmetric": True})

                if cache:
                    if not x.size1() == cache['A'].size2():
                        raise Exception(
                            "Number of variables {} does not match cached constraint matrix dimensions {}".format(
                                x.size1(), cache['A'].shape))

                    n_g_cache = cache['A'].size1()
                    n_g = g.size1()

                    if n_g_cache == n_g:
                        b = cache['b']
                        A = cache['A']
                    else:
                        g_new = g[n_g_cache:]

                        if expand:
                            expand_g_new = ca.Function('f', [x_mx], [g_new]).expand()
                            g_new = expand_g_new(x)

                        # Identify the constant term in the constraints
                        b = ca.vertcat(cache['b'],
                                       ca.substitute(g_new, x, ca.DM.zeros(x.sparsity())))

                        # Identify the linear term in the constraints
                        A = ca.vertcat(cache['A'],
                                       ca.jacobian(g_new, x))
                else:
                    if expand:
                        expand_g = ca.Function('f', [x_mx], [g]).expand()
                        g = expand_g(x)

                    # Identify the constant term in the constraints
                    b = ca.substitute(g, x, ca.DM.zeros(x.sparsity()))

                    # Identify the linear term in the constraints
                    A = ca.jacobian(g, x)

                cache['A'] = A
                cache['b'] = b

                self._solver = ca.conic("mysolver",
                                        solver_name,
                                        {
                                            'h': H.sparsity(),
                                            'a': A.sparsity()
                                        },
                                        options)
                self._solver_in = {}
                self._solver_in["h"] = ca.DM(H)
                self._solver_in["g"] = ca.DM(c)
                self._solver_in["a"] = ca.DM(A)
                self._b = ca.DM(b)

            def __call__(self, x0, lbx, ubx, lbg, ubg):

                self._solver_in["x0"] = x0
                self._solver_in["lbx"] = lbx
                self._solver_in["ubx"] = ubx
                self._solver_in["lba"] = lbg - self._b
                self._solver_in["uba"] = ubg - self._b

                solver_out = self._solver(**self._solver_in)

                solver_out['f'] = solver_out['cost']

                return solver_out

            def stats(self):
                return self._solver.stats().copy()

        return Solver()
